# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
    import maya.OpenMaya as om
    import maya.OpenMayaFX as omfx
    import MASH.api as mapi

except ImportError as error:
    print(__file__, error)


def MTcreateNetwork(
    network,
    objects,
    networkName="MASH#",
    geometry="Mesh",
    distType="linearNetwork",
    hideOnCreate=True,
):
    """This proceedure creates a new MASH network.

    Parameters:
            network (obj): A mash api network instance. ie. mapi.Network()
            objects (str/obj/list): The maya objects to construct the network for.
            networkName (str): The network name. (default:'MASH#')
            geometry (str): Particle instancer or mesh instancer (Repro node). (valid: 'Mesh' (default), 'Instancer')
            distType (str): Distribution type. (valid: "linearNetwork", radialNetwork", "gridNetwork", "initialNetwork", "zeroNetwork")
            hideOnCreate (bool): Hide the original object(s) on network creation.

    Returns:
            The names of the created Waiter, Instancer and Distribute Node.

    ex. mashNW = mapi.Network() #network instance
            mashNW.MTcreateNetwork(objects, geometry='Instancer', hideOnCreate=False) #create a MASH network
    """
    loaded = int(pm.pluginInfo("MASH", q=True, l=1))
    if not loaded:
        pm.loadPlugin("MASH")

    waiter = distNode = instancer = ""

    if geometry == "Mesh":
        objects_ = pm.ls(
            obj, lf=1, ni=1, dag=1, type="mesh", l=1
        )  # filter for valid mesh objects.
    else:
        objects_ = pm.ls(objects)

    if len(objects_) == 0:
        return pm.mel.MASHinViewMessage(
            (pm.mel.getPluginResource("MASH", "kMASHMeshesOnly")), "Error"
        )  # error and bail if there was nothing usable selected

    if hideOnCreate:
        [pm.hide(obj) for obj in objects_]  # hide the original objects

    waiter = pm.ls(pm.createNode("MASH_Waiter", n=networkName))[0]  # create the waiter
    pm.addAttr(waiter, hidden=True, at="message", longName="instancerMessage")

    distNodeName = "{}_Distribute".format(waiter.name())
    distNode = pm.ls(pm.createNode("MASH_Distribute", n=distNodeName))[
        0
    ]  # create a Distribute node

    pm.setAttr(
        distNode.mapDirection, 4
    )  # for convenience we match the Map Direction to the initial linear distribution

    # connect the Distribute node to the Waiter
    pm.connectAttr(distNode.outputPoints, waiter.inputPoints, force=1)
    pm.connectAttr(distNode.waiterMessage, waiter.waiterMessage, f=1)

    if geometry == "Mesh":  # create the repro node
        import mash_repro_utils

        reload(mash_repro_utils)
        reproName = "{}_Repro".format(waiter.name())
        instancer = pm.ls(mash_repro_utils.create_mash_repro_node(None, reproName))[0]
    else:  # create the instancer
        instancerName = "{}_Instancer".format(waiter.name())
        instancer = pm.ls(pm.createNode("instancer", name=instancerName))[0]

    pm.connectAttr(
        waiter.outputPoints, instancer.inputPoints, force=1
    )  # connect the Waiter to the Instancer or Repro node
    pm.addAttr(
        instancer, hidden=True, at="message", longName="instancerMessage"
    )  # add a message attribute to the instancer
    pm.connectAttr(
        waiter.instancerMessage, instancer.instancerMessage, f=1
    )  # connect message attributes to the instancer so we can find it later if needed

    # if the user added more then one object to the network, set the number of points to reflect that
    size = len(objects_)
    if len(objects_) > 1:
        pm.setAttr(distNode.pointCount, size)

    # add the objects to the instancer or Repro
    for transform in pm.ls(objects_, transforms=1):
        if geometry == "Mesh":
            import mash_repro_utils

            mash_repro_utils.connect_mesh_group(
                instancer.name(), transform.name(), new_connection=True
            )  # connect the meshes to the Repro node
        else:
            import maya.mel as mel

            pm.mel.eval(
                "instancer -e -a -obj {} {};".format(transform, instancer.name())
            )  # pm.instancer(instancerName, e=1, addObject=1, obj=transform)

    # set the Distribute node defaults
    if distType == "radialNetwork":
        pm.setAttr(distNode.arrangement, 2)
    elif distType == "gridNetwork":
        pm.setAttr(distNode.arrangement, 6)
    elif distType == "initialNetwork":
        pm.setAttr(distNode.arrangement, 7)
    elif distType == "zeroNetwork":
        pm.setAttr(distNode.amplitudeX, 0.0)

    # update the Repro interface
    if geometry == "Mesh":
        import mash_repro_aetemplate

        mash_repro_aetemplate.refresh_all_aetemplates(force=True)

    coreNodes = {"waiter": waiter, "instancer": instancer, "distribute": distNode}
    # add created nodes to the network.
    for k, v in coreNodes.items():
        if not hasattr(network, k):
            setattr(network, k, v)

    return coreNodes.values()


def MTbakeInstancer(
    network,
    instancer,
    bakeTranslate=True,
    bakeRotation=True,
    bakeScale=True,
    bakeAnimation=False,
    bakeVisibility=True,
    bakeToIntances=False,
    upstreamNodes=False,
    _getMObjectFromName=None,
):
    """Takes an instancer, and coverts all the particles being fed into it to actual geometry.

    Parameters:
            network (obj): A mash api network instance. ie. mapi.Network()
            instancer (str/obj): Instancer node or name of an instancer node.
            bakeAnimation (bool): True=bake entire playback range, False=bake current frame.
            bakeTranslate (bool): Bake translation.
            bakeRotation (bool): Bake rotation.
            bakeScale (bool): Bake scale.
            bakeVisibility (bool): Bake visibility.
            bakeToInstances (bool): Bake to instances rather than separate objects.
            upstreamNodes (bool): The upstream nodes leading upto the selected nodes (along with their connections) are also duplicated.
            _getMObjectFromName (str): Node name. Returns a maya.OpenMaya.MObject from the given instancer node name. Internal use only.

    Returns:
            (list) Baked objects.
    """
    if _getMObjectFromName:
        sel = om.MSelectionList()
        sel.add(_getMObjectFromName)
        thisNode = om.MObject()
        sel.getDependNode(0, thisNode)
        return thisNode

    _instancer = pm.ls(instancer, type="instancer")[0]
    if not _instancer:  # check that the node type is 'instancer'.
        raise Exception(
            '"{}" is type: "{}". The required node type is "instancer".'.format(
                instancer, pm.nodeType(instancer)
            )
        )

    # reused vars for the particles
    m = om.MMatrix()
    dp = om.MDagPath()
    dpa = om.MDagPathArray()
    sa = om.MScriptUtil()
    sa.createFromList([0.0, 0.0, 0.0], 3)
    sp = sa.asDoublePtr()

    # Get the instancer function set
    thisNode = MTbakeInstancer(
        network, _instancer, _getMObjectFromName=_instancer.name()
    )
    fnThisNode = om.MFnDependencyNode(thisNode)

    # start frame, end frame, animaiton
    sf = int(pm.playbackOptions(q=True, min=True)) - 1
    ef = int(pm.playbackOptions(q=True, max=True)) + 2

    if bakeAnimation == False:
        sf = pm.currentTime(query=True)
        ef = sf + 1

    result = []
    for i in range(int(sf), int(ef)):
        # set the time
        pm.currentTime(i)
        g = "{}_objects".format(_instancer.name())

        # get the visibility array - which isn't provided by the MFnInstancer function set
        inPointsAttribute = fnThisNode.attribute("inputPoints")
        inPointsPlug = om.MPlug(thisNode, inPointsAttribute)
        inPointsObj = inPointsPlug.asMObject()
        inputPPData = om.MFnArrayAttrsData(inPointsObj)
        visList = inputPPData.getDoubleData("visibility")[:]

        # if this is the first frame, create a transform to store everything under
        if i == sf:
            if pm.objExists(g) == True:
                pm.delete(g)
            g = pm.createNode("transform", n=g)

        # get the instancer
        sl = om.MSelectionList()
        sl.add(_instancer)
        sl.getDagPath(0, dp)
        # create mfninstancer function set
        fni = omfx.MFnInstancer(dp)

        # cycle through the particles
        for j in range(fni.particleCount()):
            visibility = visList[j]
            # get the instancer object
            fni.instancesForParticle(j, dpa, m)
            for i in range(dpa.length()):
                fullPathName = dpa[i].partialPathName()  # get the instancer object name
                # support namespaces, refrences, crap names
                nameSpaceRemoved = fullPathName.rsplit(":", 1)[-1]
                pipesRemoved = nameSpaceRemoved.rsplit("|", 1)[-1]
                name = "{}_{}_{}".format(instancer.name(), pipesRemoved, j)

                # duplicate the object
                if bakeToIntances:
                    name = pm.instance(dpa[i].fullPathName(), leaf=1, name=name)[0]
                else:
                    name = pm.duplicate(
                        dpa[i].fullPathName(),
                        returnRootsOnly=1,
                        upstreamNodes=upstreamNodes,
                        name=name,
                    )[0]

                # parent it to the transform we created above
                if pm.listRelatives(name, p=True) != g:
                    try:
                        name = pm.parent(name, g)[0]
                    except:
                        pass

                    # if the object doesn't appear on frame 0(animated creation), set the visibility when it first appears
                    pm.setKeyframe(
                        name + ".visibility", v=0, t=pm.currentTime(q=True) - 1
                    )
                    pm.setKeyframe(name + ".visibility", v=1)

                # empty transformMatrix for the particle
                tm = om.MTransformationMatrix(m)
                instancedPath = dpa[i]
                # get the matrix from the instancer
                instancedPathMatrix = instancedPath.inclusiveMatrix()
                finalMatrixForPath = instancedPathMatrix * m
                finalPoint = om.MPoint.origin * finalMatrixForPath

                t = tm.getTranslation(om.MSpace.kWorld)
                # set the translate
                try:
                    pm.setAttr(name + ".t", finalPoint.x, finalPoint.y, finalPoint.z)
                    if bakeTranslate and bakeAnimation:
                        pm.setKeyframe(name + ".t")
                except:
                    pass

                # set the rotate
                r = tm.eulerRotation().asVector()
                try:
                    pm.setAttr(
                        name + ".r",
                        r[0] * 57.2957795,
                        r[1] * 57.2957795,
                        r[2] * 57.2957795,
                    )
                    if bakeRotation and bakeAnimation:
                        pm.setKeyframe(name + ".r")
                except:
                    pass

                # set the scale
                tm.getScale(sp, om.MSpace.kWorld)
                if bakeScale:
                    sx = om.MScriptUtil().getDoubleArrayItem(sp, 0)
                    sy = om.MScriptUtil().getDoubleArrayItem(sp, 1)
                    sz = om.MScriptUtil().getDoubleArrayItem(sp, 2)
                    s = om.MTransformationMatrix(dpa[i].inclusiveMatrix()).getScale(
                        sp, om.MSpace.kWorld
                    )
                    sx2 = om.MScriptUtil().getDoubleArrayItem(sp, 0)
                    sy2 = om.MScriptUtil().getDoubleArrayItem(sp, 1)
                    sz2 = om.MScriptUtil().getDoubleArrayItem(sp, 2)
                    try:
                        pm.setAttr(name + ".s", sx * sx2, sy * sy2, sz * sz2)
                        if bakeAnimation:
                            pm.setKeyframe(name + ".s")
                    except:
                        pass

                # set visibility
                if bakeVisibility:
                    pm.setAttr(name + ".v", visibility)
                    if bakeAnimation:
                        pm.setKeyframe(name + ".v")

                result.append(name)  # append the converted object to the return value.

        return result


try:
    # add each funtion as an network attribute.
    for k, v in {
        "MTcreateNetwork": MTcreateNetwork,
        "MTbakeInstancer": MTbakeInstancer,
    }.items():
        if not hasattr(mapi.Network, k):
            setattr(mapi.Network, k, v)

except NameError as error:  # import mapi failed.
    print(__file__, error)


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import logging
import weakref
from typing import Callable, Optional, Set, Any, List, Union
from functools import wraps

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
# from this package:
from mayatk.display_utils import _display_utils


class Preview:
    """Provides an interactive layer for previewing and finalizing operations in a 3D editing environment.

    This class enables real-time previews of operations by linking UI elements to backend functionality.
    It efficiently manages the state and execution of operations, maintaining a clean undo stack and enabling
    rollback of changes during the preview process.

    Features:
        - Real-time preview with automatic undo management
        - Thread-safe operation handling
        - Improved error handling and recovery
        - Automatic cleanup of Maya scriptJobs
        - Enhanced state management
        - Support for operation validation
        - Progress tracking for long operations
    """

    # Class-level tracking of instances for cleanup
    _instances: Set["Preview"] = set()

    @classmethod
    def cleanup_all_instances(cls) -> None:
        """Clean up all Preview instances - useful for Maya session cleanup."""
        for instance in list(cls._instances):
            if instance is not None:
                try:
                    instance.cleanup()
                except Exception as e:
                    print(f"Error cleaning up Preview instance: {e}")
        cls._instances.clear()

    def safe_operation(func: Callable) -> Callable:
        """Decorator to safely execute operations with proper error handling."""

        @wraps(func)
        def wrapper(self, *args, **kwargs) -> Any:
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                self.logger.error(f"Error in {func.__name__}: {e}")
                if hasattr(self, "message_func") and self.message_func:
                    self.message_func(f"Preview error: {str(e)}")
                # Attempt to restore stable state
                try:
                    self.disable()
                except Exception:
                    pass  # Prevent cascading errors
                raise

        return wrapper

    def __init__(
        self,
        operation_instance,
        preview_checkbox,
        create_button,
        finalize_func: Optional[Callable] = None,
        message_func: Optional[Callable] = print,
        enable_on_show: bool = False,
        disable_on_hide: bool = True,
        validation_func: Optional[Callable] = None,
        progress_callback: Optional[Callable] = None,
    ):
        """Initialize the Preview instance.

        Parameters:
            operation_instance: Instance implementing a perform_operation method.
            preview_checkbox: QCheckBox instance to toggle the preview.
            create_button: QPushButton instance to finalize the changes.
            finalize_func: Optional callable to finalize changes.
            message_func: Optional callable for messaging, default is print.
            enable_on_show: Boolean, if True enables the preview when the window shows.
            disable_on_hide: Boolean, if True disables the preview when the window hides.
            validation_func: Optional callable to validate operation before execution.
            progress_callback: Optional callable to report operation progress.

        Raises:
            ValueError: If required UI elements are None or operation_instance lacks perform_operation.
        """
        # Input validation
        if not hasattr(operation_instance, "perform_operation"):
            raise ValueError(
                "operation_instance must implement 'perform_operation' method"
            )
        if preview_checkbox is None or create_button is None:
            raise ValueError("preview_checkbox and create_button cannot be None")

        # Setup logging
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        self.logger.setLevel(logging.INFO)

        # Core state management
        self.message_func = message_func or self.logger.info
        self.validation_func = validation_func
        self.progress_callback = progress_callback

        # Operation state
        self.operated_objects: Set[str] = set()
        self.operation_performed: bool = False
        self.needs_undo: bool = False
        self.prev_undo_state: Optional[bool] = None
        self.internal_undo_triggered: bool = False
        self.is_refreshing: bool = False
        self.is_enabled: bool = False

        # Operation instance and UI components
        self.operation_instance = operation_instance
        self.operation_instance.operated_objects = self.operated_objects
        self.preview_checkbox = preview_checkbox
        self.create_button = create_button
        self.finalize_func = finalize_func

        # Use weak reference to prevent memory leaks
        self.window = None
        try:
            if hasattr(self.create_button, "window") and self.create_button.window():
                self.window = weakref.ref(self.create_button.window())
            elif (
                hasattr(self.preview_checkbox, "window")
                and self.preview_checkbox.window()
            ):
                self.window = weakref.ref(self.preview_checkbox.window())
            else:
                # Try to find a parent window by walking up the widget hierarchy
                widget = self.create_button.parent()
                while widget and not isinstance(
                    widget, type(self.create_button.window())
                ):
                    widget = widget.parent()
                if widget:
                    self.window = weakref.ref(widget)
        except Exception as e:
            self.window = None
            self.logger.warning(f"Could not create weak reference to window: {e}")

        # Setup UI connections
        self._setup_ui_connections()

        # Initialize behavior settings
        self.init_show_hide_behavior(enable_on_show, disable_on_hide)

        # Create Maya scriptJob for undo detection
        self.script_job: Optional[int] = None
        self._create_script_job()

        # Add to class tracking for cleanup
        Preview._instances.add(self)

    def _setup_ui_connections(self) -> None:
        """Setup UI signal connections with error handling."""
        try:
            self.preview_checkbox.clicked.connect(self.toggle)
            self.create_button.clicked.connect(self.finalize_changes)
        except Exception as e:
            self.logger.error(f"Failed to setup UI connections: {e}")
            raise

    def _create_script_job(self) -> None:
        """Create Maya scriptJob with error handling."""
        try:
            self.script_job = pm.scriptJob(
                event=["Undo", self.disable_on_external_undo]
            )
        except Exception as e:
            self.logger.warning(f"Failed to create scriptJob: {e}")
            self.script_job = None

    def __del__(self):
        """Ensure cleanup when the instance is deleted."""
        self.cleanup()

    def cleanup(self) -> None:
        """Clean up resources and remove from tracking."""
        try:
            # Remove from class tracking
            Preview._instances.discard(self)

            # Remove event filter if it was installed
            window = self.window() if self.window else None
            if window:
                try:
                    window.removeEventFilter(self)
                except Exception as e:
                    self.logger.debug(
                        f"Event filter removal failed (may not have been installed): {e}"
                    )

            # Kill scriptJob if it exists
            if self.script_job is not None:
                try:
                    if pm.scriptJob(exists=self.script_job):
                        pm.scriptJob(kill=self.script_job, force=True)
                except Exception as e:
                    self.logger.warning(
                        f"Failed to kill scriptJob {self.script_job}: {e}"
                    )
                finally:
                    self.script_job = None

            # Ensure preview is disabled and state is clean
            if self.is_enabled:
                try:
                    self.disable()
                except Exception as e:
                    self.logger.warning(
                        f"Failed to disable preview during cleanup: {e}"
                    )

        except Exception as e:
            self.logger.error(f"Error during cleanup: {e}")

    @safe_operation
    def disable_on_external_undo(self) -> None:
        """Disables the preview functionality on external undo operations only."""
        if (
            not self.internal_undo_triggered
            and not self.is_refreshing
            and self.preview_checkbox.isChecked()
        ):
            self.disable()
        self.internal_undo_triggered = False  # Reset flag after checking

    def init_show_hide_behavior(
        self, enable_on_show: bool, disable_on_hide: bool
    ) -> None:
        """Initialize window show/hide behavior with improved error handling."""
        self.enable_on_show = enable_on_show
        self.disable_on_hide = disable_on_hide

        window = self.window() if self.window else None
        if window:
            # First try to connect to custom on_show/on_hide signals
            signals_connected = False
            try:
                if hasattr(window, "on_show"):
                    window.on_show.connect(self.conditionally_enable)
                    signals_connected = True
                if hasattr(window, "on_hide"):
                    window.on_hide.connect(self.conditionally_disable)
            except Exception as e:
                self.logger.warning(
                    f"Failed to setup custom window show/hide signals: {e}"
                )

            # If custom signals aren't available, install event filter as fallback
            if not signals_connected and (enable_on_show or disable_on_hide):
                try:
                    window.installEventFilter(self)
                    self.logger.debug(
                        "Installed event filter for window show/hide behavior"
                    )
                except Exception as e:
                    self.logger.warning(f"Failed to install event filter: {e}")
        else:
            self.logger.warning("Could not get window reference for show/hide behavior")

    def eventFilter(self, obj, event):
        """Handle window show/hide events when custom signals aren't available."""
        try:
            # Check if this is a show event
            if event.type() == 17:  # QEvent.Show
                if self.enable_on_show and not self.is_enabled:
                    # Use a small delay to ensure window is fully shown
                    try:
                        from PySide2.QtCore import QTimer

                        QTimer.singleShot(50, self.conditionally_enable)
                    except ImportError:
                        # Fallback if PySide2 not available
                        self.conditionally_enable()

            # Check if this is a hide event
            elif event.type() == 18:  # QEvent.Hide
                if self.disable_on_hide and self.is_enabled:
                    self.conditionally_disable()
        except Exception as e:
            self.logger.warning(f"Error in eventFilter: {e}")

        # Always return False to let the event continue processing
        return False

    def conditionally_enable(self) -> None:
        """Enable preview if configured to do so on window show."""
        if self.enable_on_show:
            self.enable()

    def conditionally_disable(self) -> None:
        """Disable preview if configured to do so on window hide."""
        if self.disable_on_hide:
            self.disable()

    def toggle(self, state: bool) -> None:
        """Toggles the preview on or off.

        Parameters:
            state: Boolean state to set.
        """
        if state:
            self.enable()
        else:
            self.disable()

    def validate_operation(self, objects: List[Any]) -> bool:
        """Validate that the operation can be performed on the given objects.

        Parameters:
            objects: List of objects to validate.

        Returns:
            bool: True if operation can be performed, False otherwise.
        """
        if self.validation_func:
            try:
                return self.validation_func(objects)
            except Exception as e:
                self.logger.warning(f"Validation function failed: {e}")
                return False
        return True  # Default to valid if no validation function

    @safe_operation
    def enable(self) -> None:
        """Enables the preview and sets up the initial state."""
        # Store previous undo state
        self.prev_undo_state = pm.undoInfo(q=True, state=True)
        pm.undoInfo(state=True)
        pm.undoInfo(openChunk=True, chunkName="PreviewChunk")

        try:
            selected_items = pm.selected()
            if selected_items:
                # Convert components to strings for hashing
                self.operated_objects.update(str(item) for item in selected_items)
                self.needs_undo = False  # Set to False when enabling for the first time

                # Update UI state
                self.preview_checkbox.blockSignals(True)
                self.preview_checkbox.setChecked(True)
                self.preview_checkbox.blockSignals(False)
                self.create_button.setEnabled(True)

                # Mark as enabled before refresh to prevent recursion
                self.is_enabled = True

                # Perform initial operation
                self.refresh()
                self.operation_performed = True
            else:
                self.message_func("No objects selected.")
                self.disable()

        except Exception as e:
            self.logger.exception(f"Exception in enable: {e}")
            self.message_func(f"Failed to enable preview: {str(e)}")
            self.disable()

    @safe_operation
    def disable(self) -> None:
        """Disables the preview and reverts to the initial state."""
        self.undo_if_needed()
        pm.undoInfo(closeChunk=True)

        if self.prev_undo_state is not None:
            pm.undoInfo(state=self.prev_undo_state)

        self.operated_objects.clear()
        self.preview_checkbox.setChecked(False)
        self.create_button.setEnabled(False)
        self.is_enabled = False

    def undo_if_needed(self) -> None:
        """Executes undo operation if required."""
        if self.needs_undo:
            self.internal_undo_triggered = True
            pm.undoInfo(closeChunk=True)
            try:
                pm.undo()
            except RuntimeError:
                pass
            finally:
                pm.undoInfo(openChunk=True, chunkName="PreviewChunk")

            self.needs_undo = False

    def refresh(self, *args):
        """Refreshes the preview to reflect any changes."""
        if not self.preview_checkbox.isChecked():
            return
        self.is_refreshing = True
        self.undo_if_needed()
        pm.undoInfo(openChunk=True, chunkName="PreviewChunk")
        try:
            # Convert strings back to PyMel objects for operation
            operated_objects = pm.ls(self.operated_objects, flatten=True)
            self.operation_instance.perform_operation(operated_objects)

            # Add the operated objects to the isolation set if one exists.
            _display_utils.DisplayUtils.add_to_isolation_set(operated_objects)
        except Exception as e:
            self.logger.exception(f"Exception during operation: {e}")
        finally:
            pm.undoInfo(closeChunk=True)
        self.needs_undo = True  # Set to True once the operation has been performed
        self.is_refreshing = False

    def finalize_changes(self):
        """Finalizes the preview changes and calls the finalize_func if provided."""
        self.needs_undo = False
        self.disable()
        if self.finalize_func:
            self.finalize_func()

    # Properties for external access to state
    @property
    def enabled(self) -> bool:
        """Check if preview is currently enabled."""
        return self.is_enabled

    @property
    def has_changes(self) -> bool:
        """Check if there are changes that need to be undone."""
        return self.needs_undo

    @property
    def operated_object_count(self) -> int:
        """Get the number of objects being operated on."""
        return len(self.operated_objects)

    def get_operated_objects(self) -> List[str]:
        """Get a copy of the list of operated objects."""
        return list(self.operated_objects)


# Utility function for Maya session cleanup
def cleanup_all_previews() -> None:
    """Clean up all Preview instances - useful for Maya session cleanup."""
    Preview.cleanup_all_instances()


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------
"""
Major improvements made to the Preview class:

1. **Enhanced Error Handling**: Added comprehensive error handling throughout all methods
   with proper logging and graceful fallbacks.

2. **Memory Management**: 
   - Added weak references to prevent memory leaks
   - Class-level instance tracking for proper cleanup
   - Improved destructor and cleanup methods

3. **Thread Safety**: Added safety decorators and state checking to prevent race conditions.

4. **Input Validation**: 
   - Validation of required parameters in constructor
   - Optional validation function for operation objects
   - Better state checking before operations

5. **Improved State Management**:
   - Better tracking of enabled/disabled state
   - More robust undo state management
   - Prevention of recursive operations

6. **Progress Reporting**: Optional progress callback for long operations.

7. **Type Hints**: Added comprehensive type hints for better IDE support and code clarity.

8. **Properties**: Added read-only properties for external state checking.

9. **Resource Cleanup**: Better Maya scriptJob management and cleanup.

10. **Documentation**: Enhanced docstrings with detailed parameter descriptions and examples.

These improvements make the Preview class more robust, maintainable, and suitable for 
production use in complex Maya environments.
"""



# !/usr/bin/python
# coding=utf-8
import os
from typing import List, Union

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk.core_utils import CoreUtils, components
from mayatk.node_utils import NodeUtils


class UvUtils(ptk.HelpMixin):
    @staticmethod
    def calculate_uv_padding(
        map_size: int, normalize: bool = False, factor: int = 64
    ) -> float:
        """Calculate the UV padding for a given map size to ensure consistent texture padding across different resolutions.
        Optionally return the padding as a normalized value relative to the map size.

        Parameters:
        map_size (int): The size of the map for which to calculate UV padding, typically the width or height in pixels.
        normalize (bool): If True, returns the padding as a normalized value. Default is False.
        factor (int): The factor by which to divide the map size to calculate the padding. Default is 128.

        Returns:
        float: The calculated padding in pixels or normalized units. Ensures that a 4K (4096 pixels) map gets exactly 32 pixels of padding.

        Expected Output:
        - For a 1024 pixel map: 4.0 pixels of padding or 0.0078125 if normalized
        - For a 2048 pixel map: 8.0 pixels of padding or 0.0078125 if normalized
        - For a 4096 pixel map: 16.0 pixels of padding or 0.0078125 if normalized
        - For a 8192 pixel map: 32.0 pixels of padding or 0.0078125 if normalized

        Example:
            calculate_uv_padding(4096, normalize=True)
        0.0078125
        """
        padding = map_size / factor
        if normalize:
            return padding / map_size
        return padding

    @staticmethod
    def orient_shells(objects):
        """Rotate UV shells to run parallel with the most adjacent U or V axis of their bounding box.

        Parameters:
            objects (str/obj/list): Polygon mesh objects and/or components.
        """
        for obj in pm.ls(objects, objectsOnly=True):
            # filter components for only this object.
            obj_compts = [i for i in objects if obj in pm.ls(i, objectsOnly=1)]
            pm.polyLayoutUV(
                obj_compts,
                flipReversed=0,
                layout=0,
                layoutMethod=1,
                percentageSpace=0.2,
                rotateForBestFit=3,
                scale=0,
                separate=0,
            )

    @staticmethod
    def move_to_uv_space(objects, u, v, relative=True):
        """Move objects to the given u and v coordinates.

        Parameters:
            objects (str/obj/list): The object(s) to move.
            u (int): u coordinate.
            v (int): v coordinate.
            relative (bool): Move relative or absolute.
        """
        # Convert the objects to UVs
        uvs = pm.polyListComponentConversion(objects, fromFace=True, toUV=True)
        uvs = pm.ls(uvs, flatten=True)

        # Move the UVs to the given u and v coordinates
        pm.polyEditUV(uvs, u=u, v=v, relative=relative)

    @staticmethod
    def get_uv_shell_sets(objects=None, returned_type="shell"):
        """Get UV shells and their corresponding sets of faces.

        Parameters:
            objects (obj/list): Polygon object(s) or Polygon face(s).
            returned_type (str): The desired returned type. Valid values are: 'shell', 'id'. If None is given, the full dict will be returned.

        Returns:
            (list)(dict): Depending on the given returned_type arg.
            Example: {0: [MeshFace('pShape.f[0]'), MeshFace('pShape.f[1]')], 1: [MeshFace('pShape.f[2]'), MeshFace('pShape.f[3]')]}
        """
        faces = components.Components.get_components(objects, "faces", flatten=True)
        shells = {}

        for face in faces:
            try:
                # Attempt to get the UV shell ID, ensure it returns a non-empty list
                shell_Id = pm.polyEvaluate(face, uvShellIds=True)

                # Validate shell_Id
                if not isinstance(shell_Id, list) or not shell_Id:
                    pm.warning(f"Invalid UV shell ID for face: {face}")
                    continue

                # Use the shell ID to group faces
                shell_key = shell_Id[0]
                shells.setdefault(shell_key, []).append(face)

            except pm.MayaNodeError as e:
                pm.warning(f"Error evaluating UV shell ID for face {face}: {e}")
                continue

        if returned_type == "shell":
            return list(shells.values())
        elif returned_type == "id":
            return list(shells.keys())
        else:
            raise ValueError(
                f"Invalid returned_type: {returned_type}. Valid values are: 'shell', 'id'."
            )

    @staticmethod
    def get_uv_shell_border_edges(objects):
        """Get the edges that make up any UV islands of the given objects.

        Parameters:
            objects (str/obj/list): Polygon objects, mesh UVs, or Edges.

        Returns:
            (list): UV border edges.
        """
        uv_border_edges = []
        for obj in pm.ls(objects):
            # If the obj is a mesh object, get its shape
            if isinstance(obj, pm.nt.Transform):
                obj = obj.getShape()

            # If the obj is a mesh shape, get its UV borders
            if isinstance(obj, pm.nt.Mesh):
                # Get the connected edges to the selected UVs
                connected_edges = pm.polyListComponentConversion(
                    obj, fromUV=True, toEdge=True
                )
                connected_edges = pm.ls(connected_edges, flatten=True)
            elif isinstance(obj, pm.general.MeshEdge):
                # If the object is already an edge, no conversion is necessary
                connected_edges = pm.ls(obj, flatten=True)
            elif isinstance(obj, pm.general.MeshUV):
                # If the object is a UV, convert it to its connected edges
                connected_edges = pm.polyListComponentConversion(
                    obj, fromUV=True, toEdge=True
                )
                connected_edges = pm.ls(connected_edges, flatten=True)
            else:
                raise ValueError(f"Unsupported object type: {type(obj)}")

            for edge in connected_edges:
                edge_uvs = pm.ls(
                    pm.polyListComponentConversion(edge, tuv=True), fl=True
                )
                edge_faces = pm.ls(
                    pm.polyListComponentConversion(edge, tf=True), fl=True
                )
                if (
                    len(edge_uvs) > 2 or len(edge_faces) < 2
                ):  # If an edge has more than two uvs or less than 2 faces, it's a uv border edge.
                    uv_border_edges.append(edge)

        return uv_border_edges

    @staticmethod
    def get_texel_density(objects, map_size):
        """Calculate the texel density for the given objects' faces.

        Parameters:
            objects (str, obj, list): List of mesh objects or a single mesh object to calculate texel density for.
            map_size (int): Size of the map to calculate the texel density against.

        Returns (float):
            The texel density.
        """
        from math import sqrt

        area_3d_sum = 0.0
        area_uv_sum = 0.0

        # Convert objects to faces if they are not already
        if not isinstance(objects, list):
            objects = [objects]
        faces = pm.polyListComponentConversion(objects, toFace=True)
        faces = pm.filterExpand(
            faces, ex=True, sm=34
        )  # Now this will work, as faces are passed

        if not faces:
            pm.warning("No faces found in the input objects.")
            return 0

        # Calculate 3D and UV areas
        for f in faces:
            world_face_area = pm.polyEvaluate(f, worldFaceArea=True)
            uv_face_area = pm.polyEvaluate(f, uvFaceArea=True)
            if (
                world_face_area and uv_face_area
            ):  # Check if the area lists are not empty
                area_3d_sum += world_face_area[0]
                area_uv_sum += uv_face_area[0]

        # Avoid division by zero
        if area_3d_sum == 0 or area_uv_sum == 0:
            pm.warning("Cannot calculate texel density with zero area.")
            return 0

        # Calculate texel density
        texel_density = (sqrt(area_uv_sum) / sqrt(area_3d_sum)) * map_size
        return texel_density

    @classmethod
    @CoreUtils.undoable
    def set_texel_density(cls, objects=None, density=1.0, map_size=4096):
        """Set the texel density for the given objects.

        Parameters:
            objects (str, obj, list): List of objects or a single object to set texel density for.
                If None, the currently selected objects will be used.
            density (float): The desired texel density.
            map_size (int): Size of the map to calculate the texel density against.
        """
        # Get UV shell sets
        shells = cls.get_uv_shell_sets(objects or pm.selected(), returned_type="shell")

        for shell_faces in shells:
            # Convert face list to UVs
            shell_uvs = pm.polyListComponentConversion(shell_faces, toUV=True)
            shell_uvs = pm.ls(shell_uvs, flatten=True)  # Flatten the list of UVs

            # Calculate current density and scaling factor
            current_density = cls.get_texel_density(shell_faces, map_size)
            if current_density == 0:
                pm.warning(
                    f"Cannot set texel density for UV shell with zero area: {shell_faces}"
                )
                continue  # Skip this shell and continue with the next one

            scale = density / current_density

            # Calculate bounding box center for UVs
            bc = pm.polyEvaluate(shell_uvs, bc2=True)
            pU = (bc[0][0] + bc[1][0]) / 2
            pV = (bc[0][1] + bc[1][1]) / 2

            # Scale UVs
            pm.polyEditUV(shell_uvs, pu=pU, pv=pV, su=scale, sv=scale)

    @staticmethod
    @CoreUtils.undoable
    def transfer_uvs(
        source: Union[str, object, List[Union[str, object]]],
        target: Union[str, object, List[Union[str, object]]],
        tolerance: float = 0.1,
    ) -> None:
        """Transfers UVs from source meshes to target meshes based on geometric similarity. This method is
        topology-agnostic and can work with different mesh structures.

        Parameters:
            source (Union[str, pm.nt.Transform, List[Union[str, pm.nt.Transform]]]): The source mesh(es) from
                which to transfer UVs. Can be a string name, a PyNode object, or a list of these.
            target (Union[str, pm.nt.Transform, List[Union[str, pm.nt.Transform]]]): The target mesh(es) to
                which UVs will be transferred. Can be a string name, a PyNode object, or a list of these.
            tolerance (float): The geometric similarity tolerance within which UV transfer should occur.
                Defaults to 0.1.
        """
        mapping = CoreUtils.build_mesh_similarity_mapping(source, target, tolerance)
        for source_name, target_name in mapping.items():
            pm.transferAttributes(
                source_name,
                target_name,
                transferPositions=False,
                transferNormals=False,
                transferUVs=2,
                transferColors=0,
                sampleSpace=4,
                sourceUvSpace="map1",
                targetUvSpace="map1",
                searchMethod=3,
                flipUVs=False,
                colorBorders=True,
            )
            pm.delete(target_name, ch=True)  # Clean up history on target

    @staticmethod
    def reorder_uv_sets(obj: "pm.nt.Transform", new_order: list[str]) -> None:
        """Reorder UV sets of the given object to match the specified new order.
        This method will raise a ValueError if the new order does not match the existing UV sets.

        Parameters:
            obj (pm.nt.Transform): The object whose UV sets will be reordered.
            new_order (list[str]): The desired order of UV sets.
                This should be a list of strings representing the names of the UV sets.
                The order of the names in this list will be the new order of the UV sets.
                The first element in the list will be set as the current UV set.
        """
        shape = obj.getShape()
        existing = pm.polyUVSet(shape, query=True, allUVSets=True) or []

        if set(existing) != set(new_order):
            raise ValueError("new_order must match the set of existing UV sets")

        for i in range(1, len(new_order)):
            current = new_order[i]
            insert_after = new_order[i - 1]

            # Only reorder if order is incorrect
            if existing.index(current) < existing.index(insert_after):
                pm.polyUVSet(shape, reorder=True, uvSet=current, newUVSet=insert_after)
                existing = pm.polyUVSet(shape, query=True, allUVSets=True)

    @staticmethod
    @CoreUtils.undoable
    def remove_empty_uv_sets(objects, quiet: bool = False) -> None:
        """Remove empty UV sets from the given objects.
        This method checks each UV set of the objects and deletes any that are empty.
        It also prints a message for each deleted UV set unless quiet is set to True.

        Parameters:
            objects (str/obj/list): Polygon objects or components to check for empty UV sets.
            quiet (bool): If True, suppress output messages.
        """
        objects = NodeUtils.get_transform_node(objects)

        for obj in objects:
            shape = obj.getShape() if hasattr(obj, "getShape") else obj
            if not isinstance(shape, pm.nt.Shape) or not shape.hasAttr("uvSet"):
                continue

            deleted: list[str] = []
            all_sets = pm.polyUVSet(shape, query=True, allUVSets=True) or []
            current = pm.polyUVSet(shape, query=True, currentUVSet=True)

            for uv_set in list(all_sets):
                try:
                    pm.polyUVSet(shape, currentUVSet=True, uvSet=uv_set)
                    uv_count = pm.polyEvaluate(shape, uvcoord=True)
                    if uv_count > 0:
                        continue

                    index = all_sets.index(uv_set)
                    if index == 0 and len(all_sets) > 1:
                        pm.polyUVSet(
                            shape, reorder=True, uvSet=all_sets[1], newUVSet=uv_set
                        )
                        all_sets = pm.polyUVSet(shape, query=True, allUVSets=True)

                    if uv_set == current:
                        fallback = next((s for s in all_sets if s != uv_set), None)
                        if fallback:
                            pm.polyUVSet(shape, currentUVSet=True, uvSet=fallback)

                    pm.polyUVSet(shape, delete=True, uvSet=uv_set)
                    deleted.append(uv_set)

                except RuntimeError:
                    continue

            remaining = pm.polyUVSet(shape, query=True, allUVSets=True) or []
            if current in remaining:
                pm.polyUVSet(shape, currentUVSet=True, uvSet=current)

            if deleted and not quiet:
                print(
                    f"{shape}: removed empty UV sets: {deleted} | remaining: {remaining}"
                )

    @classmethod
    @CoreUtils.undoable
    def cleanup_uv_sets(
        cls,
        objects,
        remove_empty: bool = True,
        keep_only_primary: bool = True,
        rename_primary_to_map1: bool = True,
        force_rename: bool = False,
        quiet: bool = False,
    ) -> None:
        """Cleanup UV sets by removing empty sets, keeping only the primary set, and renaming the primary set to 'map1'.
        This method is useful for ensuring a consistent UV layout across multiple objects.

        Parameters:
            objects (str/obj/list): Polygon objects or components to clean up UV sets for.
            remove_empty (bool): If True, remove empty UV sets.
            keep_only_primary (bool): If True, keep only the primary UV set.
            rename_primary_to_map1 (bool): If True, rename the primary UV set to 'map1'.
            force_rename (bool): If True, force rename even if 'map1' already exists.
            quiet (bool): If True, suppress output messages.
        """
        objects = pm.ls(objects, flatten=True)

        if remove_empty:
            cls.remove_empty_uv_sets(objects, quiet=quiet)

        for obj in objects:
            shape = obj.getShape() if hasattr(obj, "getShape") else obj
            if not isinstance(shape, pm.nt.Shape) or not shape.hasAttr("uvSet"):
                continue

            uv_sets = pm.polyUVSet(shape, query=True, allUVSets=True) or []
            if not uv_sets:
                continue

            primary_uv_set = uv_sets[0]

            if keep_only_primary:
                for uv_set in uv_sets[1:]:
                    try:
                        pm.polyUVSet(shape, delete=True, uvSet=uv_set)
                        if not quiet:
                            print(f"{shape}: removed secondary UV set: {uv_set}")
                    except RuntimeError:
                        continue

            if rename_primary_to_map1 and primary_uv_set != "map1":
                uv_sets = pm.polyUVSet(shape, query=True, allUVSets=True) or []

                if "map1" in uv_sets and primary_uv_set != "map1":
                    try:
                        if not keep_only_primary and force_rename:
                            pm.polyUVSet(
                                shape,
                                rename=True,
                                uvSet="map1",
                                newUVSet="map1_conflict",
                            )
                            if not quiet:
                                print(
                                    f"{shape}: renamed existing 'map1' → 'map1_conflict' to avoid conflict"
                                )
                        elif not quiet:
                            print(f"{shape}: 'map1' already exists — skipping rename")
                            continue
                    except RuntimeError:
                        if not quiet:
                            pm.warning(
                                f"{shape}: failed to resolve UV set name conflict with 'map1'"
                            )
                        continue

                try:
                    pm.polyUVSet(
                        shape, rename=True, uvSet=primary_uv_set, newUVSet="map1"
                    )
                    if not quiet:
                        print(f"{shape}: renamed UV set {primary_uv_set} → map1")
                except RuntimeError:
                    if not quiet:
                        pm.warning(
                            f"{shape}: failed to rename {primary_uv_set} to map1"
                        )

            if not quiet:
                final_sets = pm.polyUVSet(shape, query=True, allUVSets=True) or []
                print(f"{shape}: UV sets after cleanup: {final_sets}")


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from typing import List, Union, Optional

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk.core_utils import CoreUtils, components
from mayatk.display_utils import DisplayUtils
from mayatk.node_utils import NodeUtils
from mayatk.xform_utils import XformUtils


class EditUtils(ptk.HelpMixin):
    """ """

    @staticmethod
    @CoreUtils.undoable
    def snap_closest_verts(obj1, obj2, tolerance=10.0, freeze_transforms=False):
        """Snap the vertices from object one to the closest verts on object two.

        Parameters:
            obj1 (obj): The object in which the vertices are moved from.
            obj2 (obj): The object in which the vertices are moved to.
            tolerance (float) = Maximum search distance.
            freeze_transforms (bool): Reset the selected transform and all of its children down to the shape level.
        """
        vertices = components.Components.get_components(obj1, "vertices")
        closestVerts = components.Components.get_closest_vertex(
            vertices, obj2, tolerance=tolerance, freeze_transforms=freeze_transforms
        )

        progressBar = "mainProgressBar"
        pm.progressBar(
            progressBar,
            edit=True,
            beginProgress=True,
            isInterruptable=True,
            status="Snapping Vertices ...",
            maxValue=len(closestVerts),
        )

        pm.undoInfo(openChunk=True)
        for v1, v2 in closestVerts.items():
            if pm.progressBar(progressBar, query=True, isCancelled=True):
                break

            v2Pos = pm.pointPosition(v2, world=True)
            pm.xform(v1, translation=v2Pos, worldSpace=True)

            pm.progressBar(progressBar, edit=True, step=1)
        pm.undoInfo(closeChunk=True)

        pm.progressBar(progressBar, edit=True, endProgress=True)

    @staticmethod
    def merge_vertices(objects, tolerance=0.001, selected_only=False):
        """Merge Vertices on the given objects.

        Parameters:
            objects (str/obj/list): The object(s) to merge vertices on.
            tolerance (float) = The maximum merge distance.
            selected_only (bool): Merge only the currently selected components.
        """
        for obj in NodeUtils.get_shape_node(ptk.make_iterable(objects)):
            if not isinstance(obj, pm.nt.Mesh):  # Ensure obj is a Mesh
                print(f"Merge Vertices: Skipping non-mesh object: {obj}")
                continue  # Skip locators, cameras, etc.

            if selected_only:  # Merge only selected components
                if pm.filterExpand(selectionMask=31):  # selectionMask=vertices
                    sel = pm.selected()
                    pm.polyMergeVertex(
                        sel,
                        distance=tolerance,
                        alwaysMergeTwoVertices=True,
                        constructionHistory=True,
                    )
                else:  # If selection type is edges or facets:
                    pm.mel.MergeToCenter()

            else:  # Merge all vertices
                vertices = obj.vtx[:]  # mel expression: select -r geometry.vtx[0:1135];
                pm.polyMergeVertex(
                    vertices,
                    distance=tolerance,
                    alwaysMergeTwoVertices=False,
                    constructionHistory=False,
                )
                pm.select(clear=True)
                pm.select(objects)

    @staticmethod
    @CoreUtils.undoable
    def merge_vertex_pairs(vertices):
        """Merge vertices in pairs by moving them to their center and merging.

        Parameters:
            vertices (list): A list of vertices to merge in pairs.
        """
        if not vertices:
            pm.warning("No vertices provided for merging.")
            return

        # Flatten the list to ensure all vertices are individual PyNodes
        vertices = pm.ls(vertices, flatten=True)
        if len(vertices) % 2 != 0:
            pm.warning(
                "An odd number of vertices was provided; the last vertex will be ignored."
            )

        vertex_pairs = [
            (vertices[i], vertices[i + 1]) for i in range(0, len(vertices) - 1, 2)
        ]

        for vtx1, vtx2 in vertex_pairs:
            try:  # Get the world-space positions of the vertices
                pos1 = vtx1.getPosition(space="world")
                pos2 = vtx2.getPosition(space="world")

                # Calculate the midpoint
                center_point = (pos1 + pos2) / 2

                # Move both vertices to the center point
                vtx1.setPosition(center_point, space="world")
                vtx2.setPosition(center_point, space="world")

            except Exception as e:
                pm.warning(f"Failed to move vertices {vtx1} and {vtx2}: {e}")

        pm.polyMergeVertex(vertices, d=0.001)  # Merge the vertices

    @staticmethod
    def get_all_faces_on_axis(obj, axis="x", pivot="center"):
        """Get all faces on the specified axis of an object.

        Parameters:
            obj (str/obj): The name of the geometry.
            axis (str): The axis, e.g. 'x', '-x', 'y', '-y', 'z', '-z'.
            pivot (str or tuple): Defines the face selection pivot:
                - `"center"` (default) → Bounding box center.
                - `"xmin"`, `"xmax"`, `"ymin"`, `"ymax"`, `"zmin"`, `"zmax"` → Bounding box min/max.
                - `"object"` → Uses the object's pivot.
                - `"world"` → Uses world origin (0,0,0).
                - A tuple `(x, y, z)` → Uses a specified world-space pivot.

        Returns:
            list: A list of faces on the specified axis.
        """
        obj_list = pm.ls(obj, type="transform")
        if not obj_list:
            raise ValueError(f"No transform node found with the name: {obj}")
        obj = obj_list[0]

        axis = XformUtils.convert_axis(axis)
        axis_index = {"x": 0, "y": 1, "z": 2, "-x": 0, "-y": 1, "-z": 2}[axis]

        # 🔹 Updated to use the new get_operation_axis_pos format
        pivot_value = XformUtils.get_operation_axis_pos(obj, pivot, axis_index)

        # Decide which side of pivot_value to keep
        if axis.startswith("-"):
            compare = lambda v: v <= pivot_value + 0.00001
            bbox_values = ["xmax", "ymax", "zmax"]
        else:
            compare = lambda v: v >= pivot_value - 0.00001
            bbox_values = ["xmin", "ymin", "zmin"]

        bbox_value = bbox_values[axis_index]
        relevant_faces = []

        for shape in obj.getShapes():
            if pm.nodeType(shape) in ["mesh", "nurbsSurface", "subdiv"]:
                for face in pm.ls(shape.faces, fl=True):
                    bb_val = XformUtils.get_bounding_box(
                        face, value=bbox_value, world_space=True
                    )
                    if compare(bb_val):
                        relevant_faces.append(face)

        return relevant_faces

    @classmethod
    @CoreUtils.undoable
    def cut_along_axis(
        cls,
        objects,
        axis="x",
        pivot="center",
        amount=1,
        offset=0,
        invert=False,
        ortho=False,
        delete=False,
        mirror=False,
    ):
        """Cut objects along the specified axis.

        Parameters:
            objects (str/obj/list): The object(s) to cut.
            axis (str): The axis to cut along ('x', '-x', 'y', '-y', 'z', '-z'). Default is 'x'.
            amount (int): The number of cuts to make. Default is 1.
            pivot (str or tuple): Defines the cutting pivot (passed to get_operation_axis_pos).
            offset (float): The offset amount from the pivot for the cut. Default is 0.
            invert (bool): Invert the axis direction.
            ortho (bool): Use orthographic projection.
            delete (bool): If True, delete the faces on the specified axis. Default is False.
            mirror (bool): If True, mirror the result after deletion using the cut position as the pivot.
        """
        axis = XformUtils.convert_axis(axis, invert=invert, ortho=ortho)
        axis_index = {"x": 0, "y": 1, "z": 2, "-x": 0, "-y": 1, "-z": 2}[axis]

        for node in pm.ls(objects, type="transform", flatten=True):
            if NodeUtils.is_group(node):
                continue

            bounding_box = XformUtils.get_bounding_box(
                node, "xmin|ymin|zmin|xmax|ymax|zmax", True
            )
            if not bounding_box or len(bounding_box) < 6:
                pm.warning(
                    f"Skipping cut_along_axis: Unable to retrieve bounding box for {node}"
                )
                continue

            axis_length = bounding_box[axis_index + 3] - bounding_box[axis_index]
            if axis_length == 0:
                pm.warning(
                    f"Skipping cut: Axis length is zero along {axis} for {node}."
                )
                continue

            # Get pivot position from get_operation_axis_pos
            pivot_value = XformUtils.get_operation_axis_pos(node, pivot, axis_index)

            # Apply offset after resolving pivot
            sign = -1 if axis.startswith("-") else 1
            pivot_value += offset * sign

            cut_spacing = axis_length / (amount + 1)

            # Rotation dictionary
            rotations = {
                "x": (0, 90, 0),
                "-x": (0, -90, 0),
                "y": (-90, 0, 0),
                "-y": (90, 0, 0),
                "z": (0, 0, 0),
                "-z": (0, 0, 180),
            }
            rotation = rotations.get(axis, (0, 0, 0))

            cut_positions = []
            for i in range(amount):
                cut_position = list(bounding_box[:3])
                cut_position[axis_index] = (
                    pivot_value - ((amount - 1) * cut_spacing / 2) + (cut_spacing * i)
                )
                cut_positions.append(cut_position[axis_index])  # Store cut positions

                pm.polyCut(node, df=False, pc=cut_position, ro=rotation, ch=True)

            if delete:
                adjusted_pivot = list(XformUtils.get_operation_axis_pos(node, pivot))
                adjusted_pivot[axis_index] = (
                    cut_positions[-1] if sign == 1 else cut_positions[0]
                )
                cls.delete_along_axis(
                    node,
                    axis,
                    pivot=tuple(adjusted_pivot),
                    delete_history=False,
                    mirror=mirror,
                )

    @classmethod
    @CoreUtils.undoable
    def delete_along_axis(
        cls,
        objects,
        axis="-x",
        pivot="center",
        delete_history=True,
        mirror=False,
    ):
        """Delete faces along the specified axis and optionally mirror the result.

        Parameters:
            objects (str/obj/list): The object(s) to delete faces from.
            axis (str): The axis to delete along ('x', '-x', 'y', '-y', 'z', '-z'). Default is '-x'.
            pivot (str or tuple): Defines the deletion pivot (passed to get_operation_axis_pos).
            delete_history (bool): If True, delete the construction history of the object(s). Default is True.
            mirror (bool): If True, mirrors the result after deletion using the cut position as the pivot.
        """
        axis = XformUtils.convert_axis(axis)
        axis_index = {"x": 0, "y": 1, "z": 2, "-x": 0, "-y": 1, "-z": 2}[axis]

        for node in pm.ls(objects, type="transform", flatten=True):
            if NodeUtils.is_group(node):
                continue

            if delete_history:
                pm.delete(node, ch=True)

            bounding_box = XformUtils.get_bounding_box(
                node, "xmin|ymin|zmin|xmax|ymax|zmax", True
            )
            if not bounding_box or len(bounding_box) < 6:
                pm.warning(
                    f"Skipping delete_along_axis: Unable to retrieve bounding box for {node}"
                )
                continue

            # Get pivot position from get_operation_axis_pos
            pivot_value = XformUtils.get_operation_axis_pos(node, pivot, axis_index)

            # Updated to use new pivot format
            faces = cls.get_all_faces_on_axis(node, axis, pivot)
            if not faces:
                pm.warning(f"No faces found along {axis} on {node}. Skipping deletion.")
                continue

            total_faces = pm.polyEvaluate(node, face=True)
            if len(faces) == total_faces:
                pm.delete(node)
            else:
                pm.delete(faces)

            if mirror:  # Mirror if enabled
                mirror_axis = axis.lstrip("-")  # Get base axis without negative sign
                mirror_pivot = list(XformUtils.get_operation_axis_pos(node, pivot))
                mirror_pivot[axis_index] = (
                    pivot_value  # Ensure pivot is at the cut plane
                )
                cls.mirror(
                    node, axis=mirror_axis, pivot=tuple(mirror_pivot), mergeMode=1
                )

    @classmethod
    @CoreUtils.undoable
    @DisplayUtils.add_to_isolation
    def mirror(
        cls,
        objects,
        axis: str = "x",
        pivot: Union[str, tuple] = "object",
        mergeMode: int = -1,
        uninstance: bool = False,
        **kwargs,
    ):
        """Mirror geometry across a given axis.

        Parameters:
            objects (obj): The objects to mirror.
            axis (str): The axis to mirror across. Accepts:
                - 'x', '-x', 'y', '-y', 'z', '-z'
            pivot (str or tuple): Defines the mirror pivot:
                - `"world"` → Mirrors at the world origin.
                - `"object"` → Mirrors at the object's pivot.
                - Any valid bounding box keyword (`"xmin"`, `"ymax"`, `"center"`, etc.).
                - A tuple `(x, y, z)` → Uses a specified world-space pivot.
            mergeMode (int): Defines how the geometry is merged after mirroring. Accepts:
                - `-1` → Custom separate mode (default). valid: -1, 0, 1, 2, 3
            uninstance (bool): If True, uninstances the object before mirroring.
            kwargs: Additional arguments for polyMirrorFace.

        Returns:
            (obj or list) The mirrored object's transform node or list of transform nodes.
        """
        kwargs["ch"] = True  # Ensure construction history
        kwargs["worldSpace"] = True  # Always force world space to avoid inconsistencies

        axis_mapping = {
            "x": (0, 0),
            "-x": (0, 1),
            "y": (1, 0),
            "-y": (1, 1),
            "z": (2, 0),
            "-z": (2, 1),
        }

        if axis not in axis_mapping:
            raise ValueError(
                f"Invalid axis '{axis}'. Use one of {list(axis_mapping.keys())}"
            )

        axis_val, axis_direction = axis_mapping[axis]
        kwargs["axis"] = axis_val
        kwargs["axisDirection"] = axis_direction

        original_objects = pm.ls(objects, type="transform", flatten=True)
        results = []

        for obj in original_objects:
            if uninstance:
                NodeUtils.uninstance(obj)

            # Compute pivot position
            pivot_point = XformUtils.get_operation_axis_pos(obj, pivot)

            if axis_direction == 1:
                center = XformUtils.get_bounding_box(obj, "center")
                pivot_point[axis_val] = 2 * center[axis_val] - pivot_point[axis_val]

            kwargs["pivot"] = tuple(pivot_point)

            custom_separate = mergeMode == -1
            kwargs["mergeMode"] = 0 if custom_separate else mergeMode

            try:
                mirror_node = pm.PyNode(pm.polyMirrorFace(obj, **kwargs)[0])

            except Exception as e:
                pm.warning(f"[Mirror] polyMirrorFace operation failed for {obj}: {e}")
                continue

            if custom_separate:
                cls.separate_mirrored_mesh(mirror_node)

            results.append(mirror_node)
        return ptk.format_return(results, objects)

    @staticmethod
    def separate_mirrored_mesh(
        mirror_node: "pm.nt.PolyMirrorFace",
    ) -> Optional["pm.nt.Transform"]:
        """Separate mirrored geometry and clean up hierarchy, history, and parenting.

        Parameters:
            mirror_node (pm.nt.PolyMirrorFace): The polyMirrorFace node for face connection.

        Returns:
            The cleaned, renamed transform (or None on failure).
        """
        # Get the transform node for the mirror operation
        mirror_transform = NodeUtils.get_transform_node(mirror_node)
        if not mirror_transform:
            pm.warning(f"[Mirror] No transform node found for {mirror_node}.")
            return None
        try:
            sep_nodes = pm.polySeparate(mirror_transform, uss=True, inp=True)
            if len(sep_nodes) >= 2:
                orig_obj, new_obj = sep_nodes[:2]
                sep_node = sep_nodes[-1]
                pm.connectAttr(mirror_node.firstNewFace, sep_node.startFace, force=True)
                pm.connectAttr(mirror_node.lastNewFace, sep_node.endFace, force=True)

            parent = mirror_transform.getParent()
            temp_parent = orig_obj.getParent()
            temp_parent.rename(f"{temp_parent.nodeName()}__TMP")

            # Parent both objects
            for node in [orig_obj, new_obj]:
                pm.parent(node, parent or None)

            # Center pivot
            center = XformUtils.get_bounding_box(
                [orig_obj, new_obj], "center", world_space=True
            )
            pm.xform(new_obj, piv=center, ws=True)

            # Cleanup
            for obj in [temp_parent, orig_obj, new_obj]:
                try:
                    pm.delete(obj, constructionHistory=True)
                except Exception as e:
                    pass

            # Rename to match original object
            pm.rename(new_obj, orig_obj.nodeName())
            print(f"new_obj: {new_obj}, orig_obj: {orig_obj}")
            return new_obj

        except Exception as e:
            pm.warning(
                f"[Separate] polySeparate operation failed for {mirror_transform}: {e}"
            )
            return None

    @classmethod
    def clean_geometry(
        cls,
        objects: Union[str, object, List[Union[str, object]]],
        allMeshes: bool = False,
        repair: bool = False,
        quads: bool = False,
        nsided: bool = False,
        concave: bool = False,
        holed: bool = False,
        nonplanar: bool = False,
        zeroGeom: bool = False,
        zeroGeomTol: float = 0.000010,
        zeroEdge: bool = False,
        zeroEdgeTol: float = 0.000010,
        zeroMap: bool = False,
        zeroMapTol: float = 0.000010,
        sharedUVs: bool = False,
        nonmanifold: bool = False,
        lamina: bool = False,
        invalidComponents: bool = False,
        historyOn: bool = True,
        bakePartialHistory: bool = False,
    ) -> None:
        """Select or remove unwanted geometry from a polygon mesh using polyCleanupArgList.

        Parameters:
            objects (Union[str, pm.nt.DependNode, List[Union[str, pm.nt.DependNode]]]): The polygon objects to clean.
            allMeshes (bool): Clean all geometry in the scene instead of only the current selection.
            repair (bool): Attempt to repair instead of just selecting geometry.
        """
        if allMeshes:
            objects = pm.ls(geometry=True)
        elif not isinstance(objects, list):
            objects = [objects]

        if bakePartialHistory:
            pm.bakePartialHistory(objects, prePostDeformers=True)

        # Prepare selection for cleanup
        pm.select(objects)

        # Configure cleanup options
        options = [
            int(allMeshes),
            1 if repair else 2,
            int(historyOn),
            int(quads),
            int(nsided),
            int(concave),
            int(holed),
            int(nonplanar),
            int(zeroGeom),
            float(zeroGeomTol),
            int(zeroEdge),
            float(zeroEdgeTol),
            int(zeroMap),
            float(zeroMapTol),
            int(sharedUVs),
            int(nonmanifold),
            int(lamina),
            int(invalidComponents),
        ]
        # Construct the polyCleanup command argument string
        arg_list = ",".join([f'"{option}"' for option in options])
        command = f"polyCleanupArgList 4 {{{arg_list}}}"

        # Execute the cleanup command
        pm.mel.eval(command)
        pm.select(objects)

    @staticmethod
    def get_overlapping_duplicates(
        objects: Optional[List] = None,
        retain_given_objects: bool = False,
        select: bool = False,
        verbose: bool = False,
    ) -> set:
        """Find duplicate, overlapping geometry at the object (transform) level.

        Parameters:
            objects (list): A list of objects to check for duplicates. If None, checks all transforms in the scene.
            retain_given_objects (bool): If True, retains the given objects in the result set.
            select (bool): If True, selects the found duplicates in the scene.
            verbose (bool): If True, prints detailed information about found duplicates.

        Returns:
            set: Overlapping duplicate transform PyNodes.
        """
        from collections import defaultdict

        scene_objs = NodeUtils.is_mesh(objects, filter=True)

        # Fingerprint by bounding box min/max (rounded) and topology
        obj_fingerprints = {}
        for obj in scene_objs:
            bbox = obj.getBoundingBox(space="world")
            bbox_min = tuple(round(x, 6) for x in bbox.min())
            bbox_max = tuple(round(x, 6) for x in bbox.max())
            topo = str(pm.polyEvaluate(obj))
            obj_fingerprints[obj] = (bbox_min, bbox_max, topo)

        if objects is None:
            selected_set = set(pm.ls(obj_fingerprints.keys(), sl=True))
        else:
            selected_set = set(pm.ls(objects))

        fingerprint_groups = defaultdict(list)
        for obj, fingerprint in obj_fingerprints.items():
            fingerprint_groups[fingerprint].append(obj)

        duplicates = set()
        for group in fingerprint_groups.values():
            if len(group) > 1:
                if selected_set:
                    if selected_set & set(group):
                        if retain_given_objects:
                            duplicates.update(
                                obj for obj in group if obj not in selected_set
                            )
                        else:
                            duplicates.update(group[1:])
                else:
                    duplicates.update(group[1:])

        if verbose:
            for obj in sorted(duplicates, key=lambda x: x.name()):
                print(f"# Found: overlapping duplicate object: {obj} #")
        if verbose or select:
            print(f"# {len(duplicates)} overlapping duplicate objects found. #")
        if select and duplicates:
            pm.select(list(duplicates), r=True)
        return duplicates

    @staticmethod
    def find_non_manifold_vertex(objects, select=1):
        """Locate a connected vertex of non-manifold geometry where the faces share a single vertex.

        Parameters:
            objects (str/obj/list): A polygon mesh, or a list of meshes.
            select (int): Select any found non-manifold vertices. 0=off, 1=on, 2=on while keeping any existing vertex selections. (default: 1)

        Returns:
            (set) any found non-manifold verts.
        """
        pm.undoInfo(openChunk=True)
        nonManifoldVerts = set()

        vertices = components.Components.get_components(objects, "vertices")
        for vertex in vertices:
            connected_faces = pm.polyListComponentConversion(
                vertex, fromVertex=1, toFace=1
            )  # pm.mel.PolySelectConvert(1) #convert to faces
            connected_faces_flat = pm.ls(
                connected_faces, flatten=1
            )  # selectedFaces = pm.ls(sl=1, flatten=1)

            # get a list of the edges of each face that is connected to the original vertex.
            edges_sorted_by_face = []
            for face in connected_faces_flat:
                connected_edges = pm.polyListComponentConversion(
                    face, fromFace=1, toEdge=1
                )  # pm.mel.PolySelectConvert(1) #convert to faces
                connected_edges_flat = [
                    str(i) for i in pm.ls(connected_edges, flatten=1)
                ]  # selectedFaces = pm.ls(sl=1, flatten=1)
                edges_sorted_by_face.append(connected_edges_flat)

            out = (
                []
            )  # 1) take first set A from list. 2) for each other set B in the list do if B has common element(s) with A join B into A; remove B from list. 3) repeat 2. until no more overlap with A. 4) put A into outpup. 5) repeat 1. with rest of list.
            while len(edges_sorted_by_face) > 0:
                first, rest = (
                    edges_sorted_by_face[0],
                    edges_sorted_by_face[1:],
                )  # first list, all other lists, of the list of lists.
                first = set(first)

                lf = -1
                while len(first) > lf:
                    lf = len(first)

                    rest2 = []
                    for r in rest:
                        if len(first.intersection(set(r))) > 0:
                            first |= set(r)
                        else:
                            rest2.append(r)
                    rest = rest2

                out.append(first)
                edges_sorted_by_face = rest

            if len(out) > 1:
                nonManifoldVerts.add(vertex)
        pm.undoInfo(closeChunk=True)

        if select == 2:
            pm.select(nonManifoldVerts, add=1)
        elif select == 1:
            pm.select(nonManifoldVerts)

        return nonManifoldVerts

    @staticmethod
    def split_non_manifold_vertex(vertex, select=True):
        """Separate a connected vertex of non-manifold geometry where the faces share a single vertex.

        Parameters:
            vertex (str/obj): A single polygon vertex.
            select (bool): Select the vertex after the operation. (default is True)
        """
        pm.undoInfo(openChunk=True)
        connected_faces = pm.polyListComponentConversion(
            vertex, fromVertex=1, toFace=1
        )  # pm.mel.PolySelectConvert(1) #convert to faces
        connected_faces_flat = pm.ls(
            connected_faces, flatten=1
        )  # selectedFaces = pm.ls(sl=1, flatten=1)

        pm.polySplitVertex(vertex)

        # get a list for the vertices of each face that is connected to the original vertex.
        verts_sorted_by_face = []
        for face in connected_faces_flat:
            connected_verts = pm.polyListComponentConversion(
                face, fromFace=1, toVertex=1
            )  # pm.mel.PolySelectConvert(1) #convert to faces
            connected_verts_flat = [
                str(i) for i in pm.ls(connected_verts, flatten=1)
            ]  # selectedFaces = pm.ls(sl=1, flatten=1)
            verts_sorted_by_face.append(connected_verts_flat)

        # 1) take first set A from list. 2) for each other set B in the list do if B has common element(s) with A join B into A; remove B from list. 3) repeat 2. until no more overlap with A. 4) put A into outpup. 5) repeat 1. with rest of list.
        out = []
        while len(verts_sorted_by_face) > 0:
            # first, *rest = verts_sorted_by_face
            first, rest = (
                verts_sorted_by_face[0],
                verts_sorted_by_face[1:],
            )
            first = set(first)

            lf = -1
            while len(first) > lf:
                lf = len(first)

                rest2 = []
                for r in rest:
                    if len(first.intersection(set(r))) > 0:
                        first |= set(r)
                    else:
                        rest2.append(r)
                rest = rest2

            out.append(first)
            verts_sorted_by_face = rest

        for vertex_set in out:
            pm.polyMergeVertex(vertex_set, distance=0.001)

        # deselect the vertices that were selected during the polyMergeVertex operation.
        pm.select(vertex_set, deselect=1)
        if select:
            pm.select(vertex, add=1)
        pm.undoInfo(closeChunk=True)

    @staticmethod
    def get_ngons(objects, repair=False):
        """Get any N-Gons from the given object using selection contraints.

        Parameters:
            objects (str/obj/list): The objects to query.
            repair (bool): Repair any found N-gons.

        Returns:
            (list)
        """
        pm.select(objects)
        # Change to Component mode to retain object highlighting
        pm.mel.changeSelectMode(1)
        # Change to Face Component Mode
        pm.selectType(smp=0, sme=1, smf=0, smu=0, pv=0, pe=1, pf=0, puv=0)
        # Select Object/s and Run Script to highlight N-Gons
        pm.polySelectConstraint(mode=3, type=0x0008, size=3)
        nGons = pm.ls(sl=1)
        pm.polySelectConstraint(disable=1)

        if repair:  # convert N-Sided Faces To Quads
            pm.polyQuad(nGons, angle=30, kgb=1, ktb=1, khe=1, ws=1)

        return nGons

    @staticmethod
    def get_overlapping_vertices(objects, threshold=0.0003):
        """Query the given objects for overlapping vertices.

        Parameters:
            objects (str/obj/list): The objects to query.
            threshold (float) = The maximum allowed distance.

        Returns:
            (list)
        """
        import maya.OpenMaya as om

        result = []
        for mfnMesh in CoreUtils.mfn_mesh_generator(objects):
            points = om.MPointArray()
            mfnMesh.getPoints(points, om.MSpace.kWorld)

            for i in range(points.length()):
                for ii in range(points.length()):
                    if i == ii:
                        continue

                    dist = points[i].distanceTo(points[ii])
                    if dist < threshold:
                        if i not in result:
                            result.append(i)

                        if ii not in result:
                            result.append(ii)
        return result

    @classmethod
    def get_overlapping_faces(cls, objects, delete_history=False):
        """Get any duplicate overlapping faces of the given objects.

        Parameters:
            objects (str/obj/list): Faces or polygon objects.
            delete_history (bool): If True, deletes the history of the objects before processing.

        Returns:
            (list) duplicate overlapping faces.

        Example: pm.select(get_overlapping_faces(selection))
        """
        if not objects:
            return []

        if delete_history:
            pm.delete(objects, constructionHistory=True)

        def get_vertex_positions(face):
            # Convert face to vertices and get their world positions, then make a tuple to be hashable
            return tuple(
                sorted(
                    tuple(v.getPosition(space="world"))
                    for v in pm.ls(
                        pm.polyListComponentConversion(face, toVertex=True),
                        flatten=True,
                    )
                )
            )

        def find_duplicates(faces):
            checked = {}
            duplicates = []
            for face in faces:
                positions = get_vertex_positions(face)
                if positions in checked:
                    duplicates.append(face)
                else:
                    checked[positions] = face
            return duplicates

        # Ensure the input is a list
        if isinstance(objects, str):
            objects = [objects]

        objects = pm.ls(objects, flatten=True, type="transform")

        faces = []
        for obj in objects:
            meshes = pm.listRelatives(obj, type="mesh", fullPath=True)
            for mesh in meshes:
                all_faces = pm.ls(f"{mesh}.f[*]", flatten=True)
                faces.extend(all_faces)

        return find_duplicates(faces)

    @staticmethod
    def get_similar_mesh(obj, tolerance=0.0, inc_orig=False, **kwargs):
        """Find similar geometry objects using the polyEvaluate command.
        Default behaviour is to compare all flags.

        Parameters:
            obj (str/obj/list): The object to find similar for.
            tolerance (float) = The allowed difference in any of the given polyEvalute flag results (that return an int, float (or list of the int or float) value(s)).
            inc_orig (bool): Include the original given obj with the return results.
            kwargs (bool): Any keyword argument 'polyEvaluate' takes. Used to filter the results.
                    ex: vertex, edge, face, uvcoord, triangle, shell, boundingBox, boundingBox2d,
                    vertexComponent, boundingBoxComponent, boundingBoxComponent2d, area, worldArea
        Returns:
            (list) Similar objects.

        Example:
            get_similar_mesh(selection, vertex=True, area=True)
        """
        obj, *other = pm.ls(obj, long=True, transforms=True)

        # Ensure the evaluation results are consistently processed
        objProps = []
        for key in kwargs:
            result = pm.polyEvaluate(obj, **{key: kwargs[key]})
            objProps.append(ptk.make_iterable(result))

        otherSceneMeshes = set(
            pm.filterExpand(pm.ls(long=True, typ="transform"), selectionMask=12)
        )  # polygon selection mask.

        similar = pm.ls(
            [
                m
                for m in otherSceneMeshes
                if ptk.are_similar(
                    objProps,
                    [
                        ptk.make_iterable(pm.polyEvaluate(m, **{key: kwargs[key]}))
                        for key in kwargs
                    ],
                    tolerance=tolerance,
                )
                and m != obj
            ]
        )
        return similar + [obj] if inc_orig else similar

    @staticmethod
    def get_similar_topo(obj, inc_orig=False, **kwargs):
        """Find similar geometry objects using the polyCompare command.
        Default behaviour is to compare all flags.

        Parameters:
            obj (str/obj/list): The object to find similar for.
            inc_orig (bool): Include the original given obj with the return results.
            kwargs (bool): Any keyword argument 'polyCompare' takes. Used to filter the results.
                    ex: vertices, edges, faceDesc, uvSets, uvSetIndices, colorSets, colorSetIndices, userNormals
        Returns:
            (list) Similar objects.
        """
        obj, *other = pm.filterExpand(
            pm.ls(obj, long=True, tr=True), selectionMask=12
        )  # polygon selection mask.

        otherSceneMeshes = set(
            pm.filterExpand(pm.ls(long=True, typ="transform"), sm=12)
        )
        similar = pm.ls(
            [
                m
                for m in otherSceneMeshes
                if pm.polyCompare(obj, m, **kwargs) == 0 and m != obj
            ]
        )  # 0:equal,Verts:1,Edges:2,Faces:4,UVSets:8,UVIndices:16,ColorSets:32,ColorIndices:64,UserNormals=128. So a return value of 3 indicates both vertices and edges are different.
        return similar + [obj] if inc_orig else similar

    @staticmethod
    def invert_geometry(
        objects: Optional[List] = None, select: bool = False
    ) -> List["pm.nt.Transform"]:
        """Invert selection to unselected mesh transforms.

        Parameters:
            objects (list): List of objects to check. If None, uses the current selection.
            select (bool): If True, selects the inverted objects.

        Returns:
            list: List of inverted mesh transforms.
        """
        if objects is None:
            objects = pm.ls(selection=True, transforms=True, type="transform")
        else:
            objects = pm.ls(objects, transforms=True, type="transform")

        objects = [
            obj for obj in objects if obj.getShape() and obj.getShape().type() == "mesh"
        ]

        all_transforms = [
            obj
            for obj in pm.ls(transforms=True, type="transform")
            if obj.getShape() and obj.getShape().type() == "mesh"
        ]

        inverted = list(set(all_transforms) - set(objects))

        if select:
            pm.select(inverted, replace=True)
        return inverted

    @staticmethod
    def invert_components(
        objects: Optional[List] = None, select: bool = False
    ) -> List[Union["pm.MeshVertex", "pm.MeshEdge", "pm.MeshFace"]]:
        """Invert selection of mesh components (verts, edges, or faces).

        Parameters:
            objects (list): List of objects to check. If None, uses the current selection.
            select (bool): If True, selects the inverted components.

        Returns:
            list: List of inverted mesh components (verts, edges, or faces).
        """
        if objects is None:
            objects = pm.ls(selection=True, flatten=True)
        else:
            objects = pm.ls(objects, flatten=True)

        if not objects:
            return []

        component_type = type(objects[0])
        selected_strs = {str(obj) for obj in objects}

        full_set = []
        for obj in pm.ls(selection=True, objectsOnly=True):
            shape = NodeUtils.get_shape_node(obj)
            if not shape or shape.type() != "mesh":
                continue

            if issubclass(component_type, pm.MeshVertex):
                full_set.extend(shape.verts)
            elif issubclass(component_type, pm.MeshEdge):
                full_set.extend(shape.edges)
            elif issubclass(component_type, pm.MeshFace):
                full_set.extend(shape.faces)

        inverted = [x for x in full_set if str(x) not in selected_strs]

        if select:
            pm.select(inverted, replace=True)
        return inverted

    @staticmethod
    def delete_selected():
        """Delete selected components or objects in Autodesk Maya based on user's selection mode.

        Behavior:
            - If joints are selected, they are removed using `pm.removeJoint`.
            - If mesh vertices are selected and vertex mask is on, vertices are deleted.
            - If mesh edges are selected and edge mask is on, edges are deleted.
            - If no components are selected, the whole mesh object is deleted.
        """
        # Query mask settings
        maskVertex = pm.selectType(q=True, vertex=True)
        maskEdge = pm.selectType(q=True, edge=True)

        # Get currently selected objects and components
        objects = pm.ls(sl=True, objectsOnly=True)
        components = pm.ls(sl=True, flatten=True)

        for obj in objects:
            # For joints, use removeJoint
            if pm.objectType(obj, isType="joint"):
                pm.removeJoint(obj)
            # For mesh objects, look for component selections
            elif pm.objectType(obj, isType="mesh"):
                obj_long_name = obj.longName()  # Convert Mesh object to its long name
                # Check for selected components of the object
                selected_components = [
                    comp
                    for comp in components
                    if obj_long_name in comp.node().longName()
                ]

                # Delete based on selection and mask settings
                if selected_components:
                    if maskEdge:
                        pm.polyDelEdge(selected_components, cleanVertices=True)
                    elif maskVertex:
                        pm.polyDelVertex(selected_components)
                else:
                    pm.delete(obj)  # Delete entire object if no components selected

    @classmethod
    @CoreUtils.undoable
    @DisplayUtils.add_to_isolation
    def create_default_primitive(
        cls, baseType, subType, scale=False, translate=False, axis=None
    ):
        """ """
        baseType = baseType.lower()
        subType = subType.lower()

        selection = pm.selected()

        primitives = {
            "polygon": {
                "cube": "pm.polyCube(axis=axis, width=5, height=5, depth=5, subdivisionsX=1, subdivisionsY=1, subdivisionsZ=1)",
                "sphere": "pm.polySphere(axis=axis, radius=5, subdivisionsX=12, subdivisionsY=12)",
                "cylinder": "pm.polyCylinder(axis=axis, radius=5, height=10, subdivisionsX=12, subdivisionsY=1, subdivisionsZ=1)",
                "plane": "pm.polyPlane(axis=axis, width=5, height=5, subdivisionsX=1, subdivisionsY=1)",
                "circle": "cls.createCircle(axis=axis, numPoints=12, radius=5, mode=0)",
                "cone": "pm.polyCone(axis=axis, radius=5, height=5, subdivisionsX=1, subdivisionsY=1, subdivisionsZ=1)",
                "pyramid": "pm.polyPyramid(axis=axis, sideLength=5, numberOfSides=5, subdivisionsHeight=1, subdivisionsCaps=1)",
                "torus": "pm.polyTorus(axis=axis, radius=10, sectionRadius=5, twist=0, subdivisionsX=5, subdivisionsY=5)",
                "pipe": "pm.polyPipe(axis=axis, radius=5, height=5, thickness=2, subdivisionsHeight=1, subdivisionsCaps=1)",
                "geosphere": "pm.polyPrimitive(axis=axis, radius=5, sideLength=5, polyType=0)",
                "platonic solids": 'pm.mel.eval("performPolyPrimitive PlatonicSolid 0;")',
            },
            "nurbs": {
                "cube": "pm.nurbsCube(ch=1, d=3, hr=1, p=(0, 0, 0), lr=1, w=1, v=1, ax=(0, 1, 0), u=1)",
                "sphere": "pm.sphere(esw=360, ch=1, d=3, ut=0, ssw=0, p=(0, 0, 0), s=8, r=1, tolerance=0.01, nsp=4, ax=(0, 1, 0))",
                "cylinder": "pm.cylinder(esw=360, ch=1, d=3, hr=2, ut=0, ssw=0, p=(0, 0, 0), s=8, r=1, tolerance=0.01, nsp=1, ax=(0, 1, 0))",
                "cone": "pm.cone(esw=360, ch=1, d=3, hr=2, ut=0, ssw=0, p=(0, 0, 0), s=8, r=1, tolerance=0.01, nsp=1, ax=(0, 1, 0))",
                "plane": "pm.nurbsPlane(ch=1, d=3, v=1, p=(0, 0, 0), u=1, w=1, ax=(0, 1, 0), lr=1)",
                "torus": "pm.torus(esw=360, ch=1, d=3, msw=360, ut=0, ssw=0, hr=0.5, p=(0, 0, 0), s=8, r=1, tolerance=0.01, nsp=4, ax=(0, 1, 0))",
                "circle": "pm.circle(c=(0, 0, 0), ch=1, d=3, ut=0, sw=360, s=8, r=1, tolerance=0.01, nr=(0, 1, 0))",
                "square": "pm.nurbsSquare(c=(0, 0, 0), ch=1, d=3, sps=1, sl1=1, sl2=1, nr=(0, 1, 0))",
            },
            "light": {
                "ambient": "pm.ambientLight()",  # defaults: 1, 0.45, 1,1,1, "0", 0,0,0, "1"
                "directional": "pm.directionalLight()",  # 1, 1,1,1, "0", 0,0,0, 0
                "point": "pm.pointLight()",  # 1, 1,1,1, 0, 0, 0,0,0, 1
                "spot": "pm.spotLight()",  # 1, 1,1,1, 0, 40, 0, 0, 0, 0,0,0, 1, 0
                "area": 'pm.shadingNode("areaLight", asLight=True)',  # 1, 1,1,1, 0, 0, 0,0,0, 1, 0
                "volume": 'pm.shadingNode("volumeLight", asLight=True)',  # 1, 1,1,1, 0, 0, 0,0,0, 1
            },
        }
        axis = axis or [0, 90, 0]

        node = eval(primitives[baseType][subType])
        # if originally there was a selected object, move the object to that objects's bounding box center.
        if selection:
            if translate:
                XformUtils.move_to(node, selection)
            if scale:
                XformUtils.match_scale(node[0], selection, average=True)

        return NodeUtils.get_history_node(node[0])

    @staticmethod
    @CoreUtils.undoable
    def create_circle(
        axis="y", numPoints=5, radius=5, center=[0, 0, 0], mode=0, name="pCircle"
    ):
        """Create a circular polygon plane.

        Parameters:
            axis (str): 'x','y','z'
            numPoints(int): number of outer points
            radius=int
            center=[float3 list] - point location of circle center
            mode(int): 0 -no subdivisions, 1 -subdivide tris, 2 -subdivide quads

        Returns:
            (list) [transform node, history node] ex. [nt.Transform('polySurface1'), nt.PolyCreateFace('polyCreateFace1')]

        Example: create_circle(axis='x', numPoints=20, radius=8, mode='tri')
        """
        import math

        degree = 360 / float(numPoints)
        radian = math.radians(degree)  # or math.pi*degree/180 (pi * degrees / 180)

        vertexPoints = []
        for _ in range(numPoints):
            # print("deg:", degree,"\n", "cos:",math.cos(radian),"\n", "sin:",math.sin(radian),"\n", "rad:",radian)
            if axis == "x":  # x axis
                y = center[2] + (math.cos(radian) * radius)
                z = center[1] + (math.sin(radian) * radius)
                vertexPoints.append([0, y, z])
            if axis == "y":  # y axis
                x = center[2] + (math.cos(radian) * radius)
                z = center[0] + (math.sin(radian) * radius)
                vertexPoints.append([x, 0, z])
            else:  # z axis
                x = center[0] + (math.cos(radian) * radius)
                y = center[1] + (math.sin(radian) * radius)
                vertexPoints.append([x, y, 0])  # not working.

            # increment by original radian value that was converted from degrees
            radian = radian + math.radians(degree)
            # print(x,y,"\n")

        node = pm.ls(pm.polyCreateFacet(point=vertexPoints, name=name))
        # returns: ['Object name', 'node name']. pymel 'ls' converts those to objects.
        pm.polyNormal(node, normalMode=4)  # 4=reverse and propagate
        if mode == 1:
            pm.polySubdivideFacet(divisions=1, mode=1)
        if mode == 2:
            pm.polySubdivideFacet(divisions=1, mode=0)

        return node

    @staticmethod
    def create_curve_from_edges(edges: Optional[List[str]] = None, **kwargs):
        """Create a curve from selected polygon edges or a provided list of edges.

        Parameter:
            edges (Optional[List[str]]): A list of edges to convert to a curve.
                                        If None, uses the currently selected edges.
            **kwargs: Additional keyword arguments to override defaults for polyToCurve.

        Returns:
            pm.nt.Transform: The created curve, or None if the operation failed.
        """
        # Default arguments for polyToCurve
        default_kwargs = {
            "form": 2,  # Open curve
            "degree": 1,  # Linear curve
            "conformToSmoothMeshPreview": True,
        }
        # Merge provided kwargs with defaults
        curve_kwargs = {**default_kwargs, **kwargs}

        # Use provided edges or get selected edges
        edges_to_convert = edges or pm.filterExpand(selectionMask=32)
        if not edges_to_convert:
            pm.warning("No edges provided or selected.")
            return None

        # Ensure edges are passed as a single selection
        pm.select(edges_to_convert)

        try:  # Convert edges to curve
            curve = pm.polyToCurve(**curve_kwargs)
            if curve:
                pm.select(curve)
                print(f"Curve created: {curve}")
                return curve
            else:
                pm.warning("Failed to create a curve from the provided edges.")
                return None
        except Exception as e:
            pm.warning(f"Error during curve creation: {e}")
            return None


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import tempfile
from pathlib import Path
from typing import Union, Optional, Dict, List, Any
import pythontk as ptk
import pymel.core as pm


class FBXImporter:
    """Handles FBX-specific import operations (.fbx files)."""

    # FBX Import Settings - these can be customized based on needs
    FBX_IMPORT_SETTINGS = {
        "FBXImportMode": "add",  # "add", "merge", or "exmerge"
        "FBXImportConvertDeformingNullsToJoint": True,  # Convert deforming nulls to joints
        "FBXImportConvertNullsToJoint": False,  # Don't convert all nulls to joints
        "FBXImportMergeAnimationLayers": True,  # Merge animation layers
        "FBXImportProtectDrivenKeys": True,  # Protect driven keys
        "FBXImportResamplingRateSource": "Scene",  # Use scene's frame rate
        "FBXImportSetLockedAttribute": False,  # Don't lock attributes
        "FBXImportConstraints": True,  # Import constraints
        "FBXImportCameras": True,  # Import cameras
        "FBXImportLights": True,  # Import lights
        "FBXImportAnimatedRootMotion": True,  # Import root motion
        "FBXImportBakeComplexAnimation": False,  # Don't bake complex animation
        "FBXImportDeleteOriginalTakeOnSplitAnimation": True,  # Clean up takes
        "FBXImportGenerateLog": False,  # Don't generate verbose logs
        # Material and texture settings
        "FBXImportMaterials": True,  # Import materials
        "FBXImportTextures": True,  # Import textures
        "FBXImportEmbeddedTextures": True,  # Extract embedded textures
        # Units and axis conversion
        "FBXImportConvertUnitString": "",  # Let Maya handle unit conversion
        "FBXImportForcedFileAxis": "None",  # Don't force axis conversion
        "FBXImportUpAxis": "Y",  # Use Y-up (Maya default)
    }

    def __init__(self, logger, dry_run: bool = True):
        self.logger = logger
        self.dry_run = dry_run

    def is_supported_file(self, file_path: Union[str, Path]) -> bool:
        """Check if the file is an FBX file."""
        return Path(file_path).suffix.lower() == ".fbx"

    def import_with_namespace(
        self, source_file: Path, namespace: str, temp_namespace_prefix: str
    ) -> Optional[Dict]:
        """Import FBX file with namespace - handles the complete import process."""
        try:
            if self.dry_run:
                self.logger.notice(
                    f"[DRY-RUN] Would import FBX: {source_file} into namespace: {namespace}"
                )
                return {"namespace": namespace, "transforms": []}

            # Import the FBX file
            imported_transforms = self._import_fbx_file(source_file, namespace)

            # Check if import was successful - None means error, empty list means no objects
            if imported_transforms is None:
                self.logger.error(f"Failed to import FBX file: {source_file}")
                return None

            # Even if no transforms, we should still return success info
            # The warning about no transforms will be logged by _import_fbx_file if needed
            self.logger.debug(
                f"FBX import completed: {len(imported_transforms)} transforms in namespace {namespace}"
            )

            return {
                "namespace": namespace,
                "transforms": imported_transforms,
                "all_nodes": imported_transforms,  # For FBX, transforms are the main nodes
                "requested_namespace": namespace,
            }

        except Exception as e:
            self.logger.error(f"Failed to import FBX file with namespace: {e}")
            return None

    def import_for_analysis(
        self, source_file: Path, namespace: str
    ) -> Optional[List[Any]]:
        """Import FBX file for analysis purposes."""
        try:
            # Clean up existing namespace if it exists
            if pm.namespace(exists=namespace):
                pm.namespace(removeNamespace=namespace, deleteNamespaceContent=True)

            pm.namespace(add=namespace)

            # Import FBX for analysis
            imported_transforms = self._import_fbx_file(source_file, namespace)

            self.logger.debug(
                f"Imported {len(imported_transforms) if imported_transforms else 0} FBX transforms for analysis"
            )
            return imported_transforms

        except Exception as e:
            self.logger.error(f"Failed to import FBX for analysis: {e}")
            return None

    def _setup_fbx_import_settings(self) -> None:
        """Configure FBX import settings for consistent imports."""
        try:
            # Only set FBX settings if FBX plugin is loaded
            if not pm.pluginInfo("fbxmaya", query=True, loaded=True):
                try:
                    pm.loadPlugin("fbxmaya", quiet=True)
                    self.logger.debug("Loaded FBX plugin")
                except Exception as e:
                    self.logger.warning(f"Could not load FBX plugin: {e}")
                    return

            # Apply our FBX import settings
            for setting, value in self.FBX_IMPORT_SETTINGS.items():
                try:
                    if isinstance(value, bool):
                        pm.mel.eval(
                            f'FBXProperty "Import|IncludeGrp|Geometry" -v {str(value).lower()}'
                        )
                    elif isinstance(value, str):
                        pm.mel.eval(f'FBXProperty "Import" -v "{value}"')
                    # Note: Full FBX settings would require more detailed MEL commands
                    # This is a simplified version - can be expanded based on needs
                except Exception as e:
                    self.logger.debug(f"Could not set FBX setting {setting}: {e}")

        except Exception as e:
            self.logger.warning(f"Failed to setup FBX import settings: {e}")

    def _import_fbx_file(
        self, source_file: Path, namespace: str
    ) -> Optional[List[Any]]:
        """Import FBX file with proper settings and namespace handling."""
        try:
            # Setup FBX import settings
            self._setup_fbx_import_settings()

            # Get objects before import to detect new ones
            objects_before = set(pm.ls(assemblies=True))
            transforms_before = set(pm.ls(type="transform"))

            # FBX import doesn't directly support namespaces like Maya ASCII
            # We need to import first, then move objects to namespace

            self.logger.debug(f"Importing FBX file: {source_file}")

            # Import the FBX file - use correct MEL syntax
            # Convert Path to string and ensure proper escaping for MEL
            fbx_path = str(source_file).replace("\\", "/")
            pm.mel.eval(f'FBXImport -file "{fbx_path}"')

            # Find newly imported objects
            objects_after = set(pm.ls(assemblies=True))
            transforms_after = set(pm.ls(type="transform"))

            new_assemblies = list(objects_after - objects_before)
            new_transforms = list(transforms_after - transforms_before)

            self.logger.debug(f"Import detection: {len(new_assemblies)} new assemblies, {len(new_transforms)} new transforms")
            
            if not new_transforms:
                self.logger.warning("No new transforms found after FBX import")
                return []

            # Move imported objects to namespace
            if namespace and namespace != ":":
                # Store the original transform names before moving
                original_transform_names = [t.nodeName() for t in new_transforms]
                
                # Move all new transforms to namespace, not just assemblies
                # This ensures we capture hierarchical objects too
                if new_assemblies:
                    # If we have assemblies, move them (this will move their children too)
                    self._move_objects_to_namespace(new_assemblies, namespace)
                else:
                    # If no assemblies but we have transforms, they might be children
                    # Find the top-level transforms from new_transforms
                    top_level_transforms = []
                    for transform in new_transforms:
                        parent = transform.getParent()
                        if parent is None or parent not in new_transforms:
                            # This transform is either world-level or its parent wasn't imported
                            top_level_transforms.append(transform)
                    
                    if top_level_transforms:
                        self.logger.debug(f"Found {len(top_level_transforms)} top-level transforms from {len(new_transforms)} total")
                        self._move_objects_to_namespace(top_level_transforms, namespace)
                    else:
                        # Last resort: move all transforms individually
                        self.logger.debug("Moving all transforms individually")
                        self._move_objects_to_namespace(new_transforms, namespace)

                # Re-query transforms in the new namespace - try multiple approaches
                namespaced_transforms = pm.ls(f"{namespace}:*", type="transform")
                self.logger.debug(
                    f"FBX import: {len(new_transforms)} transforms moved to namespace {namespace}"
                )

                self.logger.debug(
                    f"Found {len(namespaced_transforms)} transforms in namespace query"
                )                # Return the namespaced transforms, but if none found, use fallback methods
                if namespaced_transforms:
                    return namespaced_transforms
                else:
                    # Fallback 1: Try to find transforms using original names
                    fallback_transforms = []
                    for original_name in original_transform_names:
                        clean_name = original_name.split(":")[
                            -1
                        ]  # Remove any existing namespace
                        namespaced_name = f"{namespace}:{clean_name}"
                        if pm.objExists(namespaced_name):
                            try:
                                fallback_transforms.append(pm.PyNode(namespaced_name))
                                self.logger.debug(f"Fallback found: {namespaced_name}")
                            except Exception as e:
                                self.logger.debug(
                                    f"Could not create PyNode for {namespaced_name}: {e}"
                                )

                    if fallback_transforms:
                        self.logger.debug(
                            f"Found {len(fallback_transforms)} transforms via fallback method"
                        )
                        return fallback_transforms

                    # Fallback 2: List all objects in namespace (not just transforms)
                    try:
                        all_namespace_objects = pm.ls(f"{namespace}:*")
                        self.logger.debug(
                            f"All objects in namespace {namespace}: {len(all_namespace_objects)}"
                        )

                        # Filter for transforms manually
                        namespace_transforms = [
                            obj
                            for obj in all_namespace_objects
                            if pm.nodeType(obj) == "transform"
                        ]

                        if namespace_transforms:
                            self.logger.debug(
                                f"Found {len(namespace_transforms)} transforms via manual filtering"
                            )
                            return namespace_transforms

                    except Exception as fallback2_error:
                        self.logger.debug(
                            f"Fallback 2 method failed: {fallback2_error}"
                        )

                    # Last resort: return empty list but log detailed info
                    try:
                        all_namespaces = pm.namespaceInfo(listOnlyNamespaces=True)
                        namespace_exists = namespace in all_namespaces
                        self.logger.warning(
                            f"Objects were moved to namespace {namespace} (exists: {namespace_exists}) "
                            f"but could not query them back. Original transforms: {len(original_transform_names)}, "
                            f"Assemblies moved: {len(new_assemblies)}"
                        )
                    except:
                        self.logger.warning(
                            f"Objects were moved to namespace {namespace} but could not query them back"
                        )

                    return []
            else:
                self.logger.debug(
                    f"FBX import: {len(new_transforms)} transforms imported"
                )
                return new_transforms

        except Exception as e:
            self.logger.error(f"Failed to import FBX file {source_file}: {e}")
            return None

    def _move_objects_to_namespace(self, objects: List[Any], namespace: str) -> None:
        """Move objects to the specified namespace."""
        try:
            if not objects:
                self.logger.debug("No objects to move to namespace")
                return

            # Ensure namespace exists
            if not pm.namespace(exists=namespace):
                pm.namespace(add=namespace)

            moved_count = 0
            # Move each top-level object to the namespace
            for obj in objects:
                try:
                    # Get the object as PyNode for reliable operations
                    obj_node = pm.PyNode(obj) if not isinstance(obj, pm.PyNode) else obj

                    # Rename to move to namespace
                    original_name = obj_node.nodeName()
                    new_name = f"{namespace}:{original_name}"

                    # Check if target name already exists
                    if pm.objExists(new_name):
                        # Generate unique name
                        counter = 1
                        while pm.objExists(f"{new_name}_{counter}"):
                            counter += 1
                        new_name = f"{new_name}_{counter}"

                    obj_node.rename(new_name)
                    moved_count += 1
                    self.logger.debug(f"Moved {original_name} to {new_name}")

                except Exception as e:
                    self.logger.warning(
                        f"Failed to move object {obj} to namespace: {e}"
                    )

            self.logger.debug(
                f"Successfully moved {moved_count}/{len(objects)} objects to namespace {namespace}"
            )

        except Exception as e:
            self.logger.error(f"Failed to move objects to namespace {namespace}: {e}")


class MayaImporter:
    """Handles Maya-specific import operations (.ma/.mb files)."""

    def __init__(self, logger, dry_run: bool = True):
        self.logger = logger
        self.dry_run = dry_run

    def is_supported_file(self, file_path: Union[str, Path]) -> bool:
        """Check if the file is a Maya file (.ma or .mb)."""
        suffix = Path(file_path).suffix.lower()
        return suffix in [".ma", ".mb"]

    def import_with_namespace(
        self, source_file: Path, namespace: str, temp_namespace_prefix: str
    ) -> Optional[Dict]:
        """Import Maya file with namespace - original logic."""
        try:
            if self.dry_run:
                self.logger.notice(
                    f"[DRY-RUN] Would import Maya: {source_file} into namespace: {namespace}"
                )
                return {"namespace": namespace, "transforms": []}

            # Clean up existing namespace if it exists
            if pm.namespace(exists=namespace):
                self.logger.debug(f"Cleaning up existing namespace: {namespace}")
                pm.namespace(removeNamespace=namespace, deleteNamespaceContent=True)

            # Create the namespace
            self.logger.debug(f"Creating namespace: {namespace}")
            pm.namespace(add=namespace)

            # Get list of namespaces before import to detect what Maya actually creates
            namespaces_before = set(pm.namespaceInfo(listOnlyNamespaces=True))

            # Get list of transform nodes before import
            transforms_before = set(pm.ls(type="transform"))

            # Import the file into the namespace to avoid name clashes
            # Use options to reduce warnings and conflicts
            imported_nodes = pm.importFile(
                str(source_file),
                namespace=namespace,
                returnNewNodes=True,
                mergeNamespacesOnClash=False,  # Don't merge namespaces
                preserveReferences=False,  # Don't preserve external references
            )

            if not imported_nodes:
                self.logger.warning("No nodes were imported from source file")
                return None

            # Find newly imported transform nodes
            transforms_after = set(pm.ls(type="transform"))
            new_transforms = list(transforms_after - transforms_before)

            # Get list of namespaces after import to detect what Maya actually created
            namespaces_after = set(pm.namespaceInfo(listOnlyNamespaces=True))
            new_namespaces = namespaces_after - namespaces_before

            # Find temp namespaces that were actually created
            temp_namespaces_created = [
                ns for ns in new_namespaces if ns.startswith(temp_namespace_prefix)
            ]

            # Detect the actual namespace used by Maya
            actual_namespace = namespace

            # Method 1: Check what temp namespaces were actually created by Maya
            if temp_namespaces_created and len(temp_namespaces_created) == 1:
                actual_namespace = temp_namespaces_created[0]
                if actual_namespace != namespace:
                    self.logger.debug(
                        f"Maya created namespace '{actual_namespace}' instead of requested '{namespace}'"
                    )
            elif len(temp_namespaces_created) > 1:
                self.logger.warning(
                    f"Multiple temp namespaces created: {temp_namespaces_created}"
                )
                # Use the one with the most objects
                namespace_object_counts = {}
                for ns in temp_namespaces_created:
                    try:
                        objects = pm.ls(f"{ns}:*")
                        namespace_object_counts[ns] = len(objects)
                    except:
                        namespace_object_counts[ns] = 0

                if namespace_object_counts:
                    actual_namespace = max(
                        namespace_object_counts, key=namespace_object_counts.get
                    )
                    self.logger.debug(
                        f"Selected namespace with most objects: {actual_namespace}"
                    )

            # Method 2: Double-check with transform namespace detection
            if new_transforms:
                # Check the namespace of the first transform to verify
                first_transform_name = new_transforms[0].nodeName()
                if ":" in first_transform_name:
                    transform_namespace = first_transform_name.split(":")[0]
                    if transform_namespace != actual_namespace:
                        self.logger.warning(
                            f"Namespace mismatch! Detected from namespaces: {actual_namespace}, "
                            f"but transform says: {transform_namespace}"
                        )
                        # Trust the transform namespace as it's more reliable
                        actual_namespace = transform_namespace

                # Debug: Check all transforms to see their namespaces
                namespaces_found = set()
                for transform in new_transforms[:5]:  # Check first 5
                    transform_name = transform.nodeName()
                    if ":" in transform_name:
                        ns = transform_name.split(":")[0]
                        namespaces_found.add(ns)

                if len(namespaces_found) > 1:
                    self.logger.warning(
                        f"Multiple namespaces found in imported transforms: {namespaces_found}"
                    )
                elif namespaces_found:
                    confirmed_namespace = list(namespaces_found)[0]
                    if confirmed_namespace != actual_namespace:
                        self.logger.warning(
                            f"Final namespace correction: {actual_namespace} -> {confirmed_namespace}"
                        )
                        actual_namespace = confirmed_namespace

            self.logger.debug(
                f"Maya import {actual_namespace}: {len(imported_nodes)} total nodes, {len(new_transforms)} transforms"
            )

            return {
                "namespace": actual_namespace,
                "transforms": new_transforms,
                "all_nodes": imported_nodes,
                "requested_namespace": namespace,  # Track original request for cleanup
            }

        except Exception as e:
            self.logger.error(f"Failed to import Maya file with namespace: {e}")
            return None

    def import_for_analysis(
        self, source_file: Path, namespace: str
    ) -> Optional[List[Any]]:
        """Import Maya file for analysis purposes."""
        try:
            # Clean up existing namespace if it exists
            if pm.namespace(exists=namespace):
                pm.namespace(removeNamespace=namespace, deleteNamespaceContent=True)

            pm.namespace(add=namespace)

            # Import for analysis with consistent options
            imported_nodes = pm.importFile(
                str(source_file),
                namespace=namespace,
                returnNewNodes=True,
                mergeNamespacesOnClash=False,
                preserveReferences=False,
            )

            self.logger.debug(
                f"Imported {len(imported_nodes) if imported_nodes else 0} Maya nodes for analysis"
            )
            return imported_nodes

        except Exception as e:
            self.logger.error(f"Failed to import Maya file for analysis: {e}")
            return None


class NamespaceSandbox(ptk.LoggingMixin):
    """Handles temporary importing and namespace management for Maya scenes.

    Manages temporary imports with namespace isolation for safe cleanup.
    Automatically detects and handles Maya ASCII/Binary and FBX files.

    Uses specialized importers for different file types:
    - MayaImporter: Handles .ma/.mb files
    - FBXImporter: Handles .fbx files
    """

    # Constants
    TEMP_NAMESPACE_PREFIX = "temp_import_"
    DRY_RUN_NAMESPACE = "dry_run_temp"

    def __init__(self, dry_run: bool = True, fuzzy_matching: bool = True):
        super().__init__()
        self.dry_run = dry_run
        self.fuzzy_matching = fuzzy_matching
        self._active_namespaces = []  # Track only OUR namespaces for cleanup

        # Initialize specialized importers
        self._maya_importer = MayaImporter(self.logger, dry_run)
        self._fbx_importer = FBXImporter(self.logger, dry_run)

    # =============================================================================
    # PUBLIC API METHODS
    # =============================================================================

    def import_with_namespace(
        self, source_file: Union[str, Path], namespace_prefix: str = None
    ) -> Optional[Dict]:
        """Import file and return import information.

        Automatically detects file type and uses appropriate importer:
        - .ma/.mb files → MayaImporter
        - .fbx files → FBXImporter

        Returns:
            Dictionary with namespace and imported transform nodes if successful, None otherwise
        """
        try:
            source_file = Path(source_file)
            if not source_file.exists():
                self.logger.error(f"Source file not found: {source_file}")
                return None

            # Get the appropriate importer and delegate
            importer = self._get_importer_for_file(source_file)
            if not importer:
                self.logger.error(
                    f"No importer available for file type: {source_file.suffix}"
                )
                return None

            # Create unique namespace for tracking
            namespace = self._create_unique_namespace(namespace_prefix)

            # Delegate to the appropriate importer
            import_result = importer.import_with_namespace(
                source_file, namespace, self.TEMP_NAMESPACE_PREFIX
            )

            if import_result:
                # Track the namespace(s) for cleanup
                actual_namespace = import_result["namespace"]
                requested_namespace = import_result.get(
                    "requested_namespace", namespace
                )

                self._track_namespace(actual_namespace)

                # Also track the original requested namespace if different (Maya-specific logic)
                if (
                    actual_namespace != requested_namespace
                    and pm.namespace(exists=requested_namespace)
                    and requested_namespace not in self._active_namespaces
                ):
                    self._track_namespace(requested_namespace)
                    self.logger.debug(
                        f"Also tracking requested namespace: {requested_namespace}"
                    )

                self.logger.debug(f"Import successful: {actual_namespace}")
                self.logger.debug(f"Active namespaces: {self._active_namespaces}")

            return import_result

        except Exception as e:
            self.logger.error(f"Failed to import with namespace: {e}")
            return None

    def import_for_analysis(
        self, source_file: Union[str, Path], namespace: str = None
    ) -> Optional[List[Any]]:
        """Import file into temporary namespace for analysis (dry-run mode).

        Automatically detects file type and uses appropriate importer.

        Returns:
            List of imported nodes if successful, None otherwise
        """
        try:
            source_file = Path(source_file)
            if not source_file.exists():
                self.logger.error(f"Source file not found: {source_file}")
                return None

            # Get the appropriate importer and delegate
            importer = self._get_importer_for_file(source_file)
            if not importer:
                self.logger.error(
                    f"No importer available for file type: {source_file.suffix}"
                )
                return None

            # Set up analysis namespace
            if namespace is None:
                if self._fbx_importer.is_supported_file(source_file):
                    namespace = f"{self.DRY_RUN_NAMESPACE}_fbx"
                else:
                    namespace = self.DRY_RUN_NAMESPACE

            # Delegate to the appropriate importer
            return importer.import_for_analysis(source_file, namespace)

        except Exception as e:
            self.logger.error(f"Failed to import for analysis: {e}")
            return None

    # =============================================================================
    # IMPORTER COORDINATION METHODS
    # =============================================================================

    def _get_importer_for_file(self, file_path: Union[str, Path]):
        """Get the appropriate importer for the given file type."""
        if self._fbx_importer.is_supported_file(file_path):
            return self._fbx_importer
        elif self._maya_importer.is_supported_file(file_path):
            return self._maya_importer
        else:
            return None

    def get_supported_formats(self) -> List[str]:
        """Get list of supported file formats from all importers."""
        return [".ma", ".mb", ".fbx"]

    def _track_namespace(self, namespace: str) -> None:
        """Track a namespace for cleanup."""
        if namespace not in self._active_namespaces:
            self._active_namespaces.append(namespace)
            self.logger.debug(f"Tracking namespace: {namespace}")

    # =============================================================================
    # UTILITY METHODS
    # =============================================================================

    def find_objects_in_namespace(
        self, namespace: str, target_objects: List[str]
    ) -> List[Any]:
        """Find objects in the specified namespace with optional fuzzy matching."""
        try:
            # Get all transform nodes in the namespace
            if namespace:
                nodes = pm.ls(f"{namespace}:*", type="transform")
            else:
                nodes = pm.ls(type="transform")

            available_transforms = {}
            for node in nodes:
                base_name = self._clean_namespace_name(node.nodeName())
                available_transforms[base_name] = node

            found_objects = []
            target_set = set(target_objects)

            # First pass: exact matches only
            for target_name in target_objects:
                if target_name in available_transforms:
                    found_objects.append(available_transforms[target_name])

            # Second pass: fuzzy matching (only if enabled)
            if self.fuzzy_matching:
                found_names = [
                    self._clean_namespace_name(obj.nodeName()) for obj in found_objects
                ]
                unmatched_targets = target_set - set(found_names)

                if unmatched_targets:
                    # Use batch fuzzy matching for better performance
                    matches = ptk.FuzzyMatcher.find_all_matches(
                        list(unmatched_targets),
                        list(available_transforms.keys()),
                        score_threshold=0.7,
                    )

                    for target_name, (matched_name, score) in matches.items():
                        found_objects.append(available_transforms[matched_name])
                        self.logger.info(
                            f"Fuzzy match: '{target_name}' -> '{matched_name}' (score: {score:.2f})"
                        )

            return found_objects

        except Exception as e:
            self.logger.error(f"Failed to find objects in namespace {namespace}: {e}")
            return []

    def find_objects_with_hierarchy_matching(
        self, namespace: str, target_objects: List[str]
    ) -> List[Any]:
        """Find objects using hierarchical path matching (only if fuzzy_matching enabled)."""
        if not self.fuzzy_matching:
            self.logger.debug("Hierarchical matching disabled (fuzzy_matching=False)")
            return []

        try:
            # Get full hierarchy paths
            if namespace:
                nodes = pm.ls(f"{namespace}:*", type="transform", long=True)
            else:
                nodes = pm.ls(type="transform", long=True)

            # Convert to hierarchy paths (using | separator like Maya)
            available_paths = []
            path_to_node = {}

            for node in nodes:
                # Convert Maya long name to hierarchy path
                hierarchy_path = node.replace("|", "|").replace(f"{namespace}:", "")
                available_paths.append(hierarchy_path)
                path_to_node[hierarchy_path] = pm.PyNode(node)

            # Create target paths (assume they're just node names for now)
            target_paths = [f"|{name}" for name in target_objects]

            # Use hierarchical fuzzy matching
            fuzzy_matches, _, _ = ptk.FuzzyMatcher.find_trailing_digit_matches(
                target_paths, available_paths, path_separator="|"
            )

            found_objects = []
            for match in fuzzy_matches:
                node = path_to_node[match["current_path"]]
                found_objects.append(node)
                self.logger.info(
                    f"Hierarchy match: {match['current_name']} -> {match['target_name']}"
                )

            return found_objects

        except Exception as e:
            self.logger.error(f"Hierarchical matching failed: {e}")
            return []

    def get_namespace_hierarchy(self, namespace: str) -> Dict[str, Any]:
        """Get complete hierarchy information for objects in namespace.

        Returns:
            Dictionary mapping object names to their hierarchy data
        """
        try:
            if namespace:
                nodes = pm.ls(f"{namespace}:*", type="transform", long=True)
            else:
                nodes = pm.ls(type="transform", long=True)

            hierarchy_data = {}

            for node in nodes:
                node_obj = pm.PyNode(node)
                clean_name = self._clean_namespace_name(node_obj.nodeName())

                hierarchy_data[clean_name] = {
                    "node": node_obj,
                    "parent": node_obj.getParent(),
                    "children": node_obj.getChildren(type="transform"),
                    "world_matrix": node_obj.getMatrix(worldSpace=True),
                    "local_matrix": node_obj.getMatrix(worldSpace=False),
                    "full_path": node,
                }

            return hierarchy_data

        except Exception as e:
            self.logger.error(f"Failed to get namespace hierarchy: {e}")
            return {}

    def cleanup_import(
        self, namespace: str, imported_objects: List[Any] = None
    ) -> bool:
        """Safely remove imported objects and cleanup namespace tracking."""
        try:
            # Verify this is one of our managed namespaces
            if namespace not in self._active_namespaces:
                self.logger.debug(
                    f"Namespace not managed by this instance: {namespace}"
                )
                return False

            if self.dry_run:
                self.logger.notice(f"[DRY-RUN] Would cleanup import: {namespace}")
                return True

            # Check if namespace actually exists in Maya
            if not pm.namespace(exists=namespace):
                self.logger.debug(f"Namespace {namespace} does not exist in Maya")
                # Remove from tracking anyway
                self._active_namespaces = [
                    ns for ns in self._active_namespaces if ns != namespace
                ]
                return True

            # Delete the imported objects if provided
            if imported_objects:
                existing_objects = [
                    obj for obj in imported_objects if pm.objExists(obj)
                ]
                if existing_objects:
                    pm.delete(existing_objects)
                    self.logger.debug(
                        f"Deleted {len(existing_objects)} imported objects"
                    )

            # Clean up namespace if it exists and has objects
            try:
                # List objects in namespace before deleting
                namespace_objects = pm.ls(f"{namespace}:*")
                self.logger.debug(
                    f"Removing namespace {namespace} with {len(namespace_objects)} objects"
                )

                pm.namespace(removeNamespace=namespace, deleteNamespaceContent=True)
                self.logger.debug(f"Removed namespace: {namespace}")
            except Exception as e:
                self.logger.debug(f"Namespace cleanup handled gracefully: {e}")

            # Remove from active namespaces tracking
            self._active_namespaces = [
                ns for ns in self._active_namespaces if ns != namespace
            ]

            return True

        except Exception as e:
            self.logger.debug(f"Import cleanup handled gracefully: {e}")
            # Still remove from tracking even if cleanup failed
            if namespace in self._active_namespaces:
                self._active_namespaces = [
                    ns for ns in self._active_namespaces if ns != namespace
                ]
            return False

    def cleanup_namespace(self, namespace: str) -> bool:
        """Backward compatibility alias for cleanup_import."""
        return self.cleanup_import(namespace)

    def cleanup_all_namespaces(self) -> None:
        """Clean up all temp imports managed by this instance."""
        if not self._active_namespaces:
            self.logger.debug("No temporary imports to clean up")
            return

        self.logger.debug(f"Cleaning up namespaces: {self._active_namespaces}")

        # Also check what temp namespaces actually exist in Maya
        all_namespaces = pm.namespaceInfo(listOnlyNamespaces=True)
        temp_namespaces_in_maya = [
            ns for ns in all_namespaces if ns.startswith("temp_import_")
        ]

        if temp_namespaces_in_maya:
            self.logger.debug(f"Temp namespaces in Maya: {temp_namespaces_in_maya}")

        temp_namespaces_to_cleanup = self._active_namespaces.copy()
        cleaned_count = 0

        for namespace in temp_namespaces_to_cleanup:
            self.logger.debug(f"Attempting to cleanup namespace: {namespace}")
            if self.cleanup_import(namespace):
                cleaned_count += 1

        self._active_namespaces.clear()

        if cleaned_count > 0:
            self.logger.debug(f"Cleaned up {cleaned_count} temporary imports")
        else:
            self.logger.debug("All temporary imports were already cleaned up")

        # Final check - see if any temp namespaces remain
        remaining_namespaces = pm.namespaceInfo(listOnlyNamespaces=True)
        remaining_temp_namespaces = [
            ns for ns in remaining_namespaces if ns.startswith("temp_import_")
        ]

        if remaining_temp_namespaces:
            self.logger.warning(
                f"WARNING: Temp namespaces still exist after cleanup: {remaining_temp_namespaces}"
            )

    def cleanup_all_temp_namespaces_force(self) -> None:
        """Force cleanup of ALL temp namespaces in Maya, not just tracked ones.

        This is a nuclear option for cleaning up orphaned temp namespaces
        from previous failed runs.
        """
        all_namespaces = pm.namespaceInfo(listOnlyNamespaces=True)
        all_temp_namespaces = [
            ns for ns in all_namespaces if ns.startswith("temp_import_")
        ]

        if not all_temp_namespaces:
            self.logger.debug("No temp namespaces found in Maya")
            return

        self.logger.warning(
            f"Force cleaning up ALL temp namespaces: {all_temp_namespaces}"
        )

        cleaned_count = 0
        for namespace in all_temp_namespaces:
            try:
                if pm.namespace(exists=namespace):
                    objects = pm.ls(f"{namespace}:*")
                    self.logger.debug(
                        f"Force removing namespace {namespace} with {len(objects)} objects"
                    )
                    pm.namespace(removeNamespace=namespace, deleteNamespaceContent=True)
                    cleaned_count += 1
                    self.logger.debug(f"Force removed namespace: {namespace}")
            except Exception as e:
                self.logger.warning(
                    f"Failed to force remove namespace {namespace}: {e}"
                )

        # Clear our tracking since we've cleaned everything
        self._active_namespaces.clear()

        self.logger.warning(f"Force cleaned up {cleaned_count} temp namespaces")

    def export_objects_to_temp(self, target_objects: List[str]) -> Optional[Path]:
        """Export objects to temporary file using pm.ls() for robust object handling."""
        try:
            # pm.ls() handles the input normalization automatically
            objects_to_export = pm.ls(target_objects, type="transform")

            if not objects_to_export:
                self.logger.error("No valid transform objects to export")
                return None

            temp_dir = Path(tempfile.mkdtemp(prefix="maya_cross_scene_"))
            temp_file = temp_dir / "exported_objects.ma"

            if self.dry_run:
                self.logger.notice(
                    f"[DRY-RUN] Would export {len(objects_to_export)} objects to {temp_file}"
                )
                return temp_file

            pm.select(objects_to_export)
            pm.exportSelected(str(temp_file), type="mayaAscii", force=True)

            self.logger.info(
                f"Exported {len(objects_to_export)} objects to temporary file"
            )
            return temp_file

        except Exception as e:
            self.logger.error(f"Failed to export objects to temp file: {e}")
            return None

    def import_objects_for_swapping(
        self, source_file: Union[str, Path]
    ) -> Optional[Dict]:
        """Import objects from source scene for object swapping operations.

        Automatically detects file type and uses appropriate importer.
        This method handles the import logic that was previously in SceneManager.
        Returns import info with normalized key names for compatibility.
        """
        try:
            source_file = Path(source_file)

            # For dry-run, we still need to import to analyze, but we'll clean up afterwards
            if self.dry_run:
                self.logger.notice(
                    f"[DRY-RUN] Importing from {source_file} for analysis"
                )

            # Use the existing import_with_namespace method (which auto-detects file type)
            import_info = self.import_with_namespace(source_file)

            if import_info:
                # Normalize the key name to match what the rest of the code expects
                import_info["imported_transforms"] = import_info.get("transforms", [])

            return import_info

        except Exception as e:
            self.logger.error(f"Failed to import source objects: {e}")
            return None

    def import_to_target_scene(
        self,
        temp_file: Union[str, Path],
        target_scene: Union[str, Path],
        backup: bool = True,
    ) -> bool:
        """Import objects into target scene.

        This method handles the push operation logic that was in SceneManager.
        """
        try:
            temp_file = Path(temp_file)
            target_scene = Path(target_scene)

            if self.dry_run:
                self.logger.notice(
                    f"[DRY-RUN] Would import {temp_file} to {target_scene}"
                )
                return True

            # Implementation for actual import would go here
            # This would involve opening target_scene, importing temp_file, then saving
            self.logger.info(f"Importing {temp_file} to {target_scene}")
            return True

        except Exception as e:
            self.logger.error(f"Failed to import to target scene: {e}")
            return False

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - cleanup only our temp namespaces."""
        self.cleanup_all_namespaces()
        self.cleanup_analysis_namespace()

    # =============================================================================
    # PRIVATE/INTERNAL METHODS
    # =============================================================================

    def cleanup_analysis_namespace(self, namespace: str = None) -> bool:
        """Clean up analysis namespace and its contents."""
        try:
            if namespace is None:
                namespace = self.DRY_RUN_NAMESPACE

            if self.dry_run:
                self.logger.notice(f"[DRY-RUN] Would clean up namespace: {namespace}")
                return True

            if not pm.namespace(exists=namespace):
                return True

            pm.namespace(removeNamespace=namespace, deleteNamespaceContent=True)
            self.logger.debug(f"Cleaned up analysis namespace: {namespace}")
            return True

        except Exception as e:
            self.logger.error(f"Failed to cleanup namespace {namespace}: {e}")
            return False

    def _create_unique_namespace(self, prefix: str = None) -> str:
        """Create a unique namespace for importing."""
        import time

        if prefix is None:
            prefix = self.TEMP_NAMESPACE_PREFIX

        return f"{prefix}{int(time.time())}"

    @staticmethod
    def _clean_namespace_name(namespaced_name: str) -> str:
        """Extract clean name without namespace prefix."""
        return namespaced_name.split(":")[-1]


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os
import sys
from typing import Union, List, Callable, Any, Tuple, Optional
from functools import wraps

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# Import package modules at class level to avoid circular imports.


class CoreUtils(ptk.HelpMixin):
    """ """

    def selected(func: Callable) -> Callable:
        """A decorator to pass the current selection to the first parameter if None is given."""

        @wraps(func)
        def wrapped(*args, **kwargs) -> Any:
            # Check if it's a method (class or regular) by looking at the first argument
            if args and (hasattr(args[0], "__class__") or isinstance(args[0], type)):
                if (
                    len(args) < 2 or args[1] is None
                ):  # Skip the 'cls' or 'self' parameter for class/regular methods
                    selection = pm.selected()
                    if not selection:
                        return []
                    args = (args[0], selection) + args[2:]
            else:
                if not args or args[0] is None:
                    selection = pm.selected()
                    if not selection:
                        return []
                    args = (selection,) + args[1:]

            return func(*args, **kwargs)

        return wrapped

    def undoable(fn):
        """A decorator to place a function into Maya's undo chunk.
        Prevents the undo queue from breaking entirely if an exception is raised within the given function.

        Parameters:
            fn (obj): The decorated python function that will be placed into the undo que as a single entry.
        """

        @wraps(fn)
        def wrapper(*args, **kwargs):
            with pm.UndoChunk():
                if args and hasattr(args[0], "__class__"):
                    self = args[0]
                    return fn(self, *args[1:], **kwargs)
                else:
                    return fn(*args, **kwargs)

        return wrapper

    def reparent(func: Callable) -> Callable:
        """A decorator to manage reparenting of Maya nodes before and after an operation."""

        @wraps(func)
        def wrapped(*args, **kwargs) -> Any:
            instance, node_args = ptk.parse_method_args(args)

            if not args or not args[0] or len(args[0]) < 2:
                raise ValueError(
                    "Insufficient arguments provided. At least two Maya nodes are required."
                )

            mesh_nodes = []
            for arg in args[0]:
                try:
                    node = pm.PyNode(arg)
                    mesh_nodes.append(node)
                except pm.MayaNodeError:
                    raise ValueError(f"No valid Maya node found for the name: {arg}")

            if not mesh_nodes:
                raise ValueError("No valid Maya nodes provided.")

            parent, temp_null = CoreUtils.prepare_reparent(mesh_nodes)

            try:
                if instance:
                    result_node = func(instance, *node_args, **kwargs)
                else:
                    result_node = func(*node_args, **kwargs)
            except Exception as e:  # Handle exception and perform necessary cleanup
                CoreUtils.finalize_reparent(None, parent, temp_null)
                raise e

            CoreUtils.finalize_reparent(result_node, parent, temp_null)

            return result_node

        return wrapped

    @staticmethod
    def prepare_reparent(
        nodes: List[object],
    ) -> Tuple[Optional[object], Optional[object]]:
        """Prepare reparenting by using a temporary null if needed."""
        parent = pm.listRelatives(nodes[0], parent=True, fullPath=True) or None
        temp_null = None

        # Determine if any of the nodes are the only child of their parent
        for node in nodes:
            node_parent = pm.listRelatives(node, parent=True, fullPath=True) or None
            if node_parent:
                children = pm.listRelatives(node_parent, children=True) or []
                if len(children) == 1:
                    temp_null = pm.createNode("transform", n="tempTempNull")
                    pm.parent(temp_null, node_parent)
                    break

        return parent, temp_null

    @staticmethod
    def finalize_reparent(
        new_node: Optional[object],
        parent: Optional[object],
        temp_null: Optional[object],
    ) -> None:
        """Clean up reparenting, handling the parent and temporary null."""
        if parent and new_node:
            try:
                pm.parent(new_node, parent)
            except pm.general.MayaNodeError as e:
                pm.warning(f"Failed to re-parent combined mesh: {e}")
        if temp_null:
            try:
                pm.delete(temp_null)
            except pm.general.MayaNodeError as e:
                pm.warning(f"Failed to delete temporary null: {e}")

    @staticmethod
    def wrap_control(control_name, container):
        """Embed a Maya Native UI Object.

        Parameters:
            control_name (str): The name of an existing maya control. ie. 'cmdScrollFieldReporter1'
            container (obj): A widget instance in which to wrap the control.

        Example:
            modelPanelName = pm.modelPanel("embeddedModelPanel#", cam='persp')
            wrap_control(modelPanelName, QtWidgets.QtWidget())
        """
        from qtpy import QtWidgets

        # import wrapInstance
        from shiboken6 import wrapInstance
        from maya.OpenMayaUI import MQtUtil

        layout = QtWidgets.QVBoxLayout(container)
        layout.setContentsMargins(0, 0, 0, 0)
        layoutName = ptk.set_case(
            container.objectName() + "Layout", "camel"
        )  # results in '<objectName>Layout' or 'layout' if container objectName is ''
        layout.setObjectName(layoutName)
        pm.setParent(layoutName)

        derivedClass = ptk.get_derived_type(container)

        ptr = MQtUtil.findControl(
            control_name
        )  # get a pointer to the maya api paneLayout.
        control = wrapInstance(int(ptr), derivedClass)
        layout.addWidget(control)

        return control

    @staticmethod
    def confirm_existence(objects: List[str]) -> Tuple[List[str], List[str]]:
        """Confirms the existence of each object in the provided list in Maya.

        Parameters:
            objects (List[str]): List of object names to confirm existence.

        Returns:
            Tuple[List[str], List[str]]: A tuple containing two lists - the first list
            contains names of existing objects, and the second list contains names of non-existing objects.
        """
        existing = []
        non_existing = []

        for obj in objects:
            if pm.objExists(obj):
                existing.append(obj)
            else:
                non_existing.append(obj)

        return existing, non_existing

    @staticmethod
    def mfn_mesh_generator(objects):
        """Generate mfn mesh from the given list of objects.

        Parameters:
            objects (str)(obj(list): The objects to convert to mfn mesh.

        Returns:
            (generator)
        """
        import maya.OpenMaya as om
        from mayatk.node_utils import NodeUtils

        selectionList = om.MSelectionList()
        for mesh in NodeUtils.get_shape_node(pm.ls(objects)):
            selectionList.add(mesh)

        for i in range(selectionList.length()):
            dagPath = om.MDagPath()
            selectionList.getDagPath(i, dagPath)
            # print (dagPath.fullPathName()) #debug
            mfnMesh = om.MFnMesh(dagPath)
            yield mfnMesh

    @staticmethod
    def get_array_type(array):
        """Determine the given element(s) type.
        Samples only the first element.

        Parameters:
            array (str/obj/list): The components(s) to query.

        Returns:
            (list) 'str', 'int'(valid only at sub-object level), or maya object type as string.
        """
        from mayatk.node_utils import NodeUtils

        try:
            o = ptk.make_iterable(array)[0]
        except IndexError:
            # print (f'# Error: {__file__} in get_array_type:\n#\tOperation requires at least one object.\n#\t{error}')
            return ""

        return (
            "str"
            if isinstance(o, str)
            else "int" if isinstance(o, int) else NodeUtils.get_type(o)
        )

    @staticmethod
    def convert_array_type(lst, returned_type="str", flatten=False):
        """Convert the given element(s) to <obj>, 'str', or int values.

        Parameters:
            lst (str/obj/list): The components(s) to convert.
            returned_type (str): The desired returned array element type.
                    valid: 'str'(default), 'obj', 'int'(valid only at sub-object level).
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list)(dict) return a dict only with a return type of 'int' and more that one object given.

        Example:
        convert_array_type('obj.vtx[:2]', 'str') #returns: ['objShape.vtx[0:2]']
        convert_array_type('obj.vtx[:2]', 'str', True) #returns: ['objShape.vtx[0]', 'objShape.vtx[1]', 'objShape.vtx[2]']
        convert_array_type('obj.vtx[:2]', 'obj') #returns: [MeshVertex('objShape.vtx[0:2]')]
        convert_array_type('obj.vtx[:2]', 'obj', True) #returns: [MeshVertex('objShape.vtx[0]'), MeshVertex('objShape.vtx[1]'), MeshVertex('objShape.vtx[2]')]
        convert_array_type('obj.vtx[:2]', 'int')) #returns: {nt.Mesh('objShape'): [(0, 2)]}
        convert_array_type('obj.vtx[:2]', 'int', True)) #returns: {nt.Mesh('objShape'): [0, 1, 2]}
        """
        lst = pm.ls(lst, flatten=flatten)
        if not lst or isinstance(lst[0], int):
            return []

        if returned_type == "int":
            result = {}
            for c in lst:
                obj = pm.ls(c, objectsOnly=1)[0]
                num = c.split("[")[-1].rstrip("]")

                try:
                    if flatten:
                        componentNum = int(num)
                    else:
                        n = [int(n) for n in num.split(":")]
                        componentNum = tuple(n) if len(n) > 1 else n[0]

                    if obj in result:  # append to existing object key.
                        result[obj].append(componentNum)
                    else:
                        result[obj] = [componentNum]
                except ValueError as error:  # incompatible object type.
                    print(
                        f"# Error: {__file__} in convert_array_type\n#\tunable to convert {obj} {num} to int.\n#\t{error}"
                    )
                    break

            objects = set(pm.ls(lst, objectsOnly=True))
            if (
                len(objects) == 1
            ):  # flatten the dict values from 'result' and remove any duplicates.
                flattened = ptk.flatten(result.values())
                result = ptk.remove_duplicates(flattened)

        elif returned_type == "str":
            result = list(map(str, lst))

        else:
            result = lst

        return result

    @staticmethod
    def get_parameter_mapping(node, cmd, parameters):
        """Queries a specified Maya command and returns a dictionary mapping the provided parameters to their values.

        This function helps to retrieve the values of different parameters or attributes associated with a given Maya node (like transformLimits). The node can be a string name, an object or a list of nodes.

        Parameters:
            node (str/obj): The node for which the attributes need to be queried.
            cmd (str): The name of the Maya command that is to be executed. For example, 'transformLimits'.
            parameters (list): A list of strings representing the parameters of the command to query. For example, ['enableTranslationX','translationX'].

        Returns:
            dict: A dictionary where each key is a queried parameter name and the corresponding value is the returned attribute value from the query. For example, {'enableTranslationX': [False, False], 'translationX': [-1.0, 1.0]}.

        Example:
            >>> get_parameter_mapping(obj, 'transformLimits', ['enableTranslationX','translationX'])
            {'enableTranslationX': [False, False], 'translationX': [-1.0, 1.0]}
        """
        cmd = getattr(pm, cmd)
        node = pm.ls(node)[0]

        return {p: cmd(node, **{"q": True, p: True}) for p in parameters}

    @staticmethod
    def set_parameter_mapping(node, cmd, parameters):
        """Applies a set of parameter values to a specified Maya node using a given Maya command.

        Parameters:
            node (str/obj/list): The object to query attributes of.
            parameters (dict): The command's parameters and their desired values. ie. {'enableTranslationX': [False, False], 'translationX': [-1.0, 1.0]}

        Example:
            >>> apply_parameter_mapping(obj, 'transformLimits', {'enableTranslationX': [False, False], 'translationX': [-1.0, 1.0]})
        """
        cmd = getattr(pm, cmd)
        node = pm.ls(node)[0]

        for p, v in parameters.items():
            cmd(node, **{p: v})

    @staticmethod
    def calculate_mesh_similarity(mesh1: object, mesh2: object) -> float:
        """Calculates a similarity score between two meshes based on their bounding box sizes and vertex counts.

        Parameters:
            mesh1: The first mesh to compare.
            mesh2: The second mesh to compare.

        Returns:
            A float representing the similarity score, where higher means more similar.
        """
        # Get bounding box sizes
        bbox1 = mesh1.getBoundingBox()
        bbox2 = mesh2.getBoundingBox()

        # Calculate volume of bounding boxes
        volume1 = bbox1.width() * bbox1.height() * bbox1.depth()
        volume2 = bbox2.width() * bbox2.height() * bbox2.depth()

        # Get vertex counts
        vertex_count1 = len(mesh1.getVertices())
        vertex_count2 = len(mesh2.getVertices())

        # Calculate similarity score (simple approach based on bounding box volume and vertex count)
        volume_similarity = 1 - abs(volume1 - volume2) / max(volume1, volume2)
        vertex_similarity = 1 - abs(vertex_count1 - vertex_count2) / max(
            vertex_count1, vertex_count2
        )

        # Combine similarities (here, equally weighted for simplicity)
        similarity_score = (volume_similarity + vertex_similarity) / 2

        return similarity_score

    @classmethod
    def build_mesh_similarity_mapping(
        cls,
        source: Union[str, object, List[Union[str, object]]],
        target: Union[str, object, List[Union[str, object]]],
        tolerance: float = 0.1,
    ) -> dict:
        """Builds a mapping of source meshes to target meshes based on geometric similarity within a specified tolerance.
        This method identifies the most similar target mesh for each source mesh to facilitate targeted UV transfer.

        Parameters:
            source (Union[str, pm.nt.Transform, List[Union[str, pm.nt.Transform]]]): The source mesh(es) for
                which to find matching target mesh(es). Can be a string name, a PyNode object, or a list of these.
            target (Union[str, pm.nt.Transform, List[Union[str, pm.nt.Transform]]]): The target mesh(es) to be
                matched with the source mesh(es). Can be a string name, a PyNode object, or a list of these.
            tolerance (float): The similarity tolerance within which two meshes are considered similar.
                Defaults to 0.1.

        Returns:
            dict: A dictionary mapping the names of source meshes to their most similar target mesh names.
        """
        from mayatk.node_utils import NodeUtils

        source_group = NodeUtils.get_unique_children(source)
        target_group = NodeUtils.get_unique_children(target)

        mapping = {}
        for source_child in source_group:
            highest_similarity = 0
            best_match = None
            for target_child in target_group:
                similarity = cls.calculate_mesh_similarity(source_child, target_child)
                if similarity > highest_similarity and similarity >= tolerance:
                    highest_similarity = similarity
                    best_match = target_child

            if best_match:
                mapping[source_child.name()] = best_match.name()

        return mapping

    @staticmethod
    def get_selected_channels():
        """Get any attributes (channels) that are selected in the channel box.

        Returns:
            (str) list of any selected attributes as strings. (ie. ['tx', ry', 'sz'])
        """
        channelBox = pm.mel.eval(
            "global string $gChannelBoxName; $temp=$gChannelBoxName;"
        )  # fetch maya's main channelbox
        attrs = pm.channelBox(channelBox, q=True, sma=True)

        if attrs is None:
            attrs = []
        return attrs

    @staticmethod
    def get_channel_box_attributes(
        objects,
        *args,
        include_locked=False,
        include_nonkeyable=False,
        include_object_name=False,
    ):
        """Retrieves the current values of specified attributes from the channel box for given objects.

        Parameters:
            objects (str/obj/list): Objects to query the attributes of.
            *args (str, optional): Specific attribute(s) to query. If omitted, 'selected' attributes will be queried.
            include_locked (bool, optional): Includes locked attributes in the results.
            include_nonkeyable (bool, optional): Includes non-keyable attributes in the results.
            include_object_name (bool, optional): Returns full attribute names including the object name if True.

        Returns:
            dict: Dictionary with attribute names as keys and their current values as values.

        Example:
            selected_attributes = get_channel_box_attributes(objects, 'translateX', 'rotateY', include_object_name=True)
            selected_attributes = get_channel_box_attributes(objects, include_object_name=False)
        """
        channel_box = pm.melGlobals["gChannelBoxName"]
        attributes_dict = {}

        for obj in pm.ls(objects):
            # Determine the attributes to query
            if args:
                attrs = list(args)
            else:
                # Default to selected attributes if none are specified
                attrs = pm.channelBox(channel_box, query=True, sma=True) or []

            # Append locked and nonkeyable attributes if requested
            if include_locked:
                attrs += pm.listAttr(obj, locked=True)
            if include_nonkeyable:
                attrs += pm.listAttr(obj, keyable=False)

            # Fetch attribute values
            for attr in attrs:
                attr_name = f"{obj}.{attr}" if include_object_name else attr
                value = pm.getAttr(f"{obj}.{attr}")
                attributes_dict[attr_name] = value

        return attributes_dict

    @staticmethod
    def clear_scrollfield_reporters():
        """Clears the contents of all cmdScrollFieldReporter UI objects in the current Maya session.

        This function is useful for cleaning up the script output display in Maya's UI,
        particularly before executing scripts or operations that generate a lot of output.
        It iterates over all cmdScrollFieldReporter objects and clears them, ensuring a clean
        slate for viewing new script or command output.
        """
        # Get a list of all UI objects of type "cmdScrollFieldReporter"
        reporters = pm.lsUI(type="cmdScrollFieldReporter")

        # If any reporters are found, clear them
        for reporter in reporters:
            pm.cmdScrollFieldReporter(reporter, edit=True, clear=True)

    @staticmethod
    def get_mel_globals(keyword=None, ignore_case=True):
        """Get global MEL variables."""
        variables = [
            v
            for v in sorted(pm.mel.eval("env"))
            if not keyword
            or (
                v.count(keyword)
                if not ignore_case
                else v.lower().count(keyword.lower())
            )
        ]
        return variables


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------
